"""Wrappers for the models defined in redback.

https://github.com/nikhil-sarin/redback
https://redback.readthedocs.io/en/latest/
"""

import astropy.units as uu
from citation_compass import CiteClass, cite_inline

from lightcurvelynx.astro_utils.unit_utils import flam_to_fnu
from lightcurvelynx.math_nodes.bilby_priors import BilbyPriorNode
from lightcurvelynx.models.physical_model import SEDModel


class RedbackWrapperModel(SEDModel, CiteClass):
    """A wrapper for redback models.

    Parameterized values include:
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The t0 of the zero phase, date. [from BasePhysicalModel]
    Additional parameterized values are used for specific redback models.

    References
    ----------
    * redback - https://ui.adsabs.harvard.edu/abs/2024MNRAS.531.1203S/abstract
    * Individual models might require citation. See references in the redback documentation.

    Attributes
    ----------
    source : function
        The underlying source function that maps time + wavelength to flux.
    source_name : str
        The name used to set the source.
    source_param_names : list
        A list of the source model's parameters that we need to set.

    Parameters
    ----------
    source : str or function
        The name of the redback model function used to generate the SEDs or
        the actual function itself.
    priors : dict, bilby.prior.PriorDict, or BilbyPriorNode, optional
        The redback model's Bilby priors.
    parameters : dict, optional
        A dictionary of parameter setters to pass to the source function.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    # A class variable for the units so we are not computing them each time.
    _FLAM_UNIT = uu.erg / uu.second / uu.cm**2 / uu.AA

    def __init__(
        self,
        source,
        *,
        priors=None,
        parameters=None,
        **kwargs,
    ):
        # Check that the parameters passed in the dictionary and keyword arguments
        # do not overlap, so we only have one source of truth. This is needed for
        # parameters like `redshift` that overlap core parameters.
        if parameters is None:
            parameters = {}
        for key in parameters:
            if key in kwargs:
                raise ValueError(
                    f"Parameter '{key}' specified in both the parameters dictionary "
                    "and as a parameter itself. Please include it only in the dictionary."
                )

        super().__init__(**kwargs)

        # Add all of the items from the bilby prior node as settable parameters.
        parameters = parameters.copy()
        if priors is not None:
            if not isinstance(priors, BilbyPriorNode):
                priors = BilbyPriorNode(prior=priors)
            for param_name in priors.outputs:
                if param_name not in parameters:
                    parameters[param_name] = getattr(priors, param_name)

        # Use the parameter dictionary to create settable parameters for the model.
        # Some of these might have already been added by the superclass's constructor,
        # so we just change it.
        self.source_param_names = []
        for key, value in parameters.items():
            if key in self.setters:
                self.set_parameter(key, value)
            else:
                self.add_parameter(key, value, description="Parameter for redback model.")
            self.source_param_names.append(key)

        # Create the source itself.
        if isinstance(source, str):
            try:
                import redback
            except ImportError as err:  # pragma: no cover
                raise ImportError(
                    "redback package is not installed be default. To use the RedbackWrapperModel, "
                    "please install redback. For example, you can install it with "
                    "`pip install redback`."
                ) from err

            self.source_name = source
            if source not in redback.model_library.all_models_dict:  # pragma: no cover
                raise ValueError(f"Redback model '{source}' not found in redback model library.")
            self.source = redback.model_library.all_models_dict[source]
        else:
            self.source_name = source.__name__
            self.source = source

        # Check if the model has a citation parameter we should include.
        if hasattr(self.source, "citation"):
            cite_inline("redback model", self.source.citation)

        # Redback models already handle redshift, so we do not want to double apply it.
        self.apply_redshift = False

        # We save a cached version of the last computed SED. This starts as None
        # since we have not computed any SEDs yet.
        self._cached_data = {}

    @property
    def param_names(self):
        """Return a list of the model's parameter names."""
        return self.source_param_names

    def minwave(self, graph_state=None):
        """Get the minimum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. Not used
            for this model.

        Returns
        -------
        minwave : float or None
            The minimum wavelength of the model (in angstroms) or None
            if the model does not have a defined minimum wavelength.
        """
        return self._cached_data.get("minwave", None)

    def maxwave(self, graph_state=None):
        """Get the maximum wavelength of the model.

        Parameters
        ----------
        graph_state : GraphState, optional
            An object mapping graph parameters to their values. Not used
            for this model.

        Returns
        -------
        maxwave : float or None
            The maximum wavelength of the model (in angstroms) or None
            if the model does not have a defined maximum wavelength.
        """
        return self._cached_data.get("maxwave", None)

    def minphase(self, **kwargs):
        """Get the minimum supported phase of the model in days.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        minphase : float or None
            The minimum phase of the model (in days) or None
            if the model does not have a defined minimum phase.
        """
        return self._cached_data.get("minphase", None)

    def maxphase(self, **kwargs):
        """Get the maximum supported phase of the model in days.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        maximum : float or None
            The maximum phase of the model (in days) or None
            if the model does not have a defined maximum phase.
        """
        return self._cached_data.get("maxphase", None)

    def _do_bounds_precomputation(self, times, wavelengths, graph_state=None, **kwargs):
        """Precompute the SED model for the given times and wavelengths.

        This is needed for redback models because bounds [minwave, maxwave] and
        [minphase, maxphase] depend on the last computed SED.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps (MJD).
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
           Any additional keyword arguments.
        """
        # Check that the cached data is empty.
        if len(self._cached_data) != 0:  # pragma: no cover
            raise RuntimeError(
                "Cached data should be empty before precomputation. This indicates a bug "
                " in the code where the cached data is not being cleared after use."
            )

        # Build the function arguments from the parameter values.
        params = self.get_local_params(graph_state)
        fn_args = {}
        for name in self.source_param_names:
            fn_args[name] = params[name]

        # Compute the shifted times.
        t0 = params.get("t0", 0.0)
        if t0 is None:
            t0 = 0.0
        shifted_times = times - t0

        # Call the source function to get the RedbackTimeSeriesSource object.
        # We create this object with each call, because it depends on the parameters (fn_args).
        rb_result = self.source(
            shifted_times,
            output_format="sncosmo_source",
            **fn_args,
        )

        # Save the computed RedbackTimeSeriesSource and the bounds.
        self._cached_data["minwave"] = rb_result.minwave()
        self._cached_data["maxwave"] = rb_result.maxwave()
        self._cached_data["minphase"] = rb_result.minphase()
        self._cached_data["maxphase"] = rb_result.maxphase()
        self._cached_data["last_sed"] = rb_result

    def compute_sed_with_extrapolation(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object, extrapolating
        to times and wavelengths where the model is not defined.

        We override this method because the extrapolation bounds will depend on the materialized
        RedbackTimeSeriesSource object, so we need to do the precomputation to get that object.
        We will cache the object so we don't perform the computation twice.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        # Do any precomputation that is needed to set the bounds. This generates
        # cached data that will be used in the compute_sed method to avoid redundant
        # computation.
        self._do_bounds_precomputation(times, wavelengths, graph_state, **kwargs)

        # Call the superclass method to do the extrapolation. This will call compute_sed,
        # which will use the cached data to avoid redundant computation.
        sed = super().compute_sed_with_extrapolation(times, wavelengths, graph_state, **kwargs)

        # Clear the cached data values.
        self._cached_data.clear()

        return sed

    def compute_sed(self, times, wavelengths, graph_state=None, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps (MJD).
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        if self._cached_data.get("last_sed") is None:
            # If we have not computed the result object for the current parameters, do so now.
            self._do_bounds_precomputation(times, wavelengths, graph_state, **kwargs)
            created_cache = True
        else:
            created_cache = False

        # Load the RedbackTimeSeriesSource we need for this computation.
        rb_result = self._cached_data["last_sed"]

        # Compute the shifted times. Note these might be different from the times used in
        # precomputation because of the applied phase bounds.
        # times used for precomputation.
        params = self.get_local_params(graph_state)
        t0 = params.get("t0", 0.0)
        if t0 is None:
            t0 = 0.0
        shifted_times = times - t0

        # Query the RedbackTimeSeriesSource at the given times and wavelengths.
        # Then convert the output to nJy.
        model_flam = rb_result.get_flux_density(shifted_times, wavelengths)
        model_fnu = flam_to_fnu(
            model_flam,
            wavelengths,
            wave_unit=uu.AA,
            flam_unit=self._FLAM_UNIT,
            fnu_unit=uu.nJy,
        )

        # Clear the cached data values if we created them locally.
        if created_cache:
            self._cached_data.clear()

        return model_fnu



"""Test the PyLIMAWrapperModel. Note this test only runs if pylima is installed
(which is it not by default)."""

from importlib.util import find_spec

import numpy as np
import pytest
from lightcurvelynx.models.pylima_models import PyLIMAWrapperModel

if find_spec("pyLIMA") is None:
    pytest.skip("pyLIMA not installed, skipping tests.", allow_module_level=True)  # type: ignore


def test_pylima_create():
    """Test that we can create and query a PyLIMA model."""
    source_mags = {
        "g": 22.0 + np.random.normal(scale=0.3),
        "r": 21.5 + np.random.normal(scale=0.3),
        "i": 21.2 + np.random.normal(scale=0.3),
    }
    # No i in blend mags. It gets defaulted to very faint.
    blend_mags = {
        "g": 24.5,
        "r": 24.0,
    }
    pylima_params = {
        "u0": 0.01,
        "tE": 25.0,
        "piEN": 0.1,
        "piEE": 0.1,
    }

    model = PyLIMAWrapperModel(
        "PSPL",
        source_mags=source_mags,
        blend_mags=blend_mags,
        ra=1.0,
        dec=-20.0,
        t0=64350.0,
        pylima_params=pylima_params,
        blend_flux_parameter="fblend",
        node_label="source",
    )

    state = model.sample_parameters()
    assert state["source"]["u0"] == pytest.approx(0.01)
    assert state["source"]["tE"] == pytest.approx(25.0)
    assert state["source"]["piEN"] == pytest.approx(0.1)
    assert state["source"]["piEE"] == pytest.approx(0.1)
    for band in source_mags:
        assert isinstance(state["source"][f"fsource_{band}"], float)
        assert isinstance(state["source"][f"fblend_{band}"], float)

    query_times = 64350.0 + np.arange(-50, 50, 1)
    query_filters = np.array(["g"] * len(query_times))
    fluxes = model.evaluate_bandfluxes(None, query_times, query_filters, state)
    assert np.all(np.diff(fluxes[0:49]) > 0.0)  # Rising
    assert np.all(np.diff(fluxes[51:]) < 0.0)  # Falling


def test_pylima_fail_create():
    """Test that we can create and query a PyLIMA model."""
    source_mags = {
        "g": 22.0 + np.random.normal(scale=0.3),
        "r": 21.5 + np.random.normal(scale=0.3),
        "i": 21.2 + np.random.normal(scale=0.3),
    }
    # tE is missing
    pylima_params = {
        "u0": 0.01,
        "piEN": 0.1,
        "piEE": 0.1,
    }

    with pytest.raises(ValueError):
        _ = PyLIMAWrapperModel(
            "PSPL",
            source_mags=source_mags,
            blend_mags={},
            ra=1.0,
            dec=-20.0,
            t0=64350.0,
            pylima_params=pylima_params,
            blend_flux_parameter="fblend",
            node_label="source",
        )



"""The Spectrograph object stores information about a spectrograph's bins
and provides methods to compute fluxes for each bin.
"""

import numpy as np


class Spectrograph:
    """Models all of the bins of a spectrograph, producing bandfluxes for each
    bin in the spectra. This class operates similarly to a PassbandGroup, but
    only contains a single "filter" named "spectra" that contains all of the bins.

    Attributes
    ----------
    waves : np.ndarray
        The wavelengths at the center of each bin in Angstroms.
    instrument : str
        The instrument name for the spectrograph. Default is "Spectrograph".
    scale : np.ndarray
        The multiplicative factor to apply to each bin's flux to capture sensor
        sensitivity, etc. If None, we use 1.0 for all bins.
    wave_min : float
        The minimum wavelength of the spectra in Angstroms.
    wave_max : float
        The maximum wavelength of the spectra in Angstroms.
    """

    def __init__(
        self,
        waves: np.array,
        *,
        scale: np.ndarray | None = None,
        instrument: str | None = None,
    ):
        if np.any(np.diff(waves) <= 0):
            raise ValueError("waves must be in strictly increasing order.")
        self.waves = waves
        self.instrument = instrument if instrument is not None else "Spectrograph"

        self.wave_min = self.waves[0] - 0.5 * self.bin_width(0)
        self.wave_max = self.waves[-1] + 0.5 * self.bin_width(len(self.waves) - 1)

        # Scale is the multiplicative factor to apply to each bin's flux. If None, we use 1.0 for all bins.
        if scale is None:
            scale = np.ones(len(self.waves))
        elif len(scale) != len(self.waves):
            raise ValueError("Scale array must have the same length as the number of bins in the spectra.")
        self.scale = scale

    def __str__(self) -> str:
        """Return a string representation of the spectra filter."""
        return f"{self.instrument} (spectra) [{self.wave_min}A - {self.wave_max}A]"

    def __len__(self) -> int:
        return len(self.waves)

    def __eq__(self, other) -> bool:
        """Determine if two passbands have equal values for the processed tables."""
        if len(self.waves) != len(other.waves):
            return False
        if not np.allclose(self.waves, other.waves):
            return False
        if not np.allclose(self.scale, other.scale):
            return False
        return True

    @classmethod
    def from_regular_grid(cls, wave_start: float, wave_end: float, bin_width: float, **kwargs):
        """Create a Spectrograph with regularly spaced bins.

        Parameters
        ----------
        wave_start : float
            The starting wavelength of the spectra in Angstroms.
        wave_end : float
            The ending wavelength of the spectra in Angstroms.
        bin_width : float
            The bin size of the spectra in Angstroms.
        **kwargs
            Additional keyword arguments to pass to the Spectrograph constructor.

        Returns
        -------
        Spectrograph
            A Spectrograph object with regularly spaced bins.
        """
        if wave_end <= wave_start:
            raise ValueError("wave_end must be greater than wave_start.")
        if bin_width <= 0:
            raise ValueError("bin_width must be positive.")

        # We use the wavelength at the center of each bin.
        bin_centers = np.arange(wave_start + bin_width / 2, wave_end, bin_width)
        return cls(bin_centers, **kwargs)

    def bin_width(self, index):
        """Get the width of the bin at the given index.

        Parameters
        ----------
        index : int
            The index of the bin.

        Returns
        -------
        float
            The width of the bin in Angstroms.
        """
        if index < 0 or index >= len(self.waves):
            raise IndexError(f"Index {index} out of bounds for {len(self.waves)} bin widths.")
        elif index == 0:
            # The center of the bin is at waves[0], so we assume the lower bound is symmetric
            # about that point: 2.0 * (center_1 - center_0) / 2.0
            return self.waves[1] - self.waves[0]
        elif index == len(self.waves) - 1:
            # The center of the bin is at waves[-1], so we assume the upper bound is symmetric
            # about that point 2.0 * (center_N-1 - center_N-2) / 2.0
            return self.waves[-1] - self.waves[-2]
        else:
            # Half the distance to the neighboring bins on either side.
            return (self.waves[index + 1] - self.waves[index - 1]) / 2

    def wave_bounds(self):
        """Get the minimum and maximum wavelength for this spectra.

        Returns
        -------
        min_wave : float
            The minimum wavelength.
        max_wave : float
            The maximum wavelength.
        """
        return self.wave_min, self.wave_max

    def evaluate(
        self,
        flux_density_matrix: np.ndarray,
    ) -> np.ndarray:
        """Calculate the measured values for each bin in the spectrograph.

        Parameters
        ----------
        flux_density_matrix : np.ndarray
            A 2D or 3D array of flux densities. If the array is 2D it contains a single sample where
            the rows are the T times and columns are M wavelengths in Angstroms. If the array is 3D
            it contains S samples and the values are indexed as (sample_num, time, wavelength).

        Returns
        -------
        measured_flux : np.ndarray
            A 2D or 3D array. If the flux_density_matrix contains a single sample (2D input) then
            the function returns a 2D matrix where each row is a time and each column is the measurement
            at the corresponding wavelength bin. Otherwise the function returns a size S x T x B array
            where each entry corresponds to the measured value for a given sample at a given time and
            wavelength bin.
        """
        if flux_density_matrix.size == 0:
            raise ValueError("Empty flux density matrix used.")  # pragma: no cover
        if len(flux_density_matrix.shape) == 2:
            return flux_density_matrix * self.scale[np.newaxis, :]
        elif len(flux_density_matrix.shape) == 3:
            return flux_density_matrix * self.scale[np.newaxis, np.newaxis, :]
        else:
            raise ValueError("Invalid flux density matrix. Must be 2 or 3-dimensional.")  # pragma: no cover



"""Test magnitude-flux conversion utilities."""

import numpy as np
import pytest
from lightcurvelynx.astro_utils.mag_flux import (
    Flux2MagNode,
    Mag2FluxNode,
    flux2mag,
    mag2flux,
)
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc


def test_flux2mag():
    """Test that mag2flux is correct."""
    flux = np.array([3631e9, 1e9, 3631])
    desired_mag = np.array([0, 8.9, 22.5])
    np.testing.assert_allclose(flux2mag(flux), desired_mag, atol=1e-3)


def test_mag2flux():
    """Test that flux2mag is correct."""
    mag = np.array([0, 8.9, 8.9 + 2.5 * 9])
    desired_flux = np.array([3631e9, 1e9, 1])
    np.testing.assert_allclose(mag2flux(mag), desired_flux, rtol=1e-3)


def test_mag2flux2mag():
    """Test that mag2flux inverts flux2mag."""
    rng = np.random.default_rng(42)
    mag = rng.uniform(-10, 30, 1024)
    flux = mag2flux(mag)
    mag2 = flux2mag(flux)
    np.testing.assert_allclose(mag, mag2, rtol=1e-10)


def test_flux2mag2flux():
    """Test that flux2mag inverts mag2flux."""
    rng = np.random.default_rng(43)
    flux = rng.uniform(1e-3, 1e3, 1024)
    mag = flux2mag(flux)
    flux2 = mag2flux(mag)
    np.testing.assert_allclose(flux, flux2, rtol=1e-10)


def test_flux2magnode():
    """Test the computation of the Flux2MagNode."""
    fluxes = np.array([3631e9, 1e9, 3631])
    expected = flux2mag(fluxes)

    for idx, f in enumerate(fluxes):
        node = Flux2MagNode(flux_njy=f)
        state = node.sample_parameters(num_samples=1)
        assert node.get_param(state, "function_node_result") == pytest.approx(expected[idx])


def test_mag2fluxnode():
    """Test the computation of the Mag2FluxNode."""
    mags = np.array([0, 8.9, 8.9 + 2.5 * 9])
    expected = mag2flux(mags)

    for idx, m in enumerate(mags):
        node = Mag2FluxNode(mag=m)
        state = node.sample_parameters(num_samples=1)
        assert node.get_param(state, "function_node_result") == pytest.approx(expected[idx])


def test_flux2magnode_chained():
    """Test chaining Flux2MagNode and Mag2FluxNode."""
    flux_node = NumpyRandomFunc("uniform", low=100.0, high=1e6, seed=101, node_label="node1")
    mag_node = Flux2MagNode(flux_njy=flux_node, node_label="node2")

    num_samples = 10
    state_flux = mag_node.sample_parameters(num_samples=num_samples)
    assert np.allclose(
        state_flux["node2"]["function_node_result"],
        flux2mag(state_flux["node1"]["function_node_result"]),
    )



"""A basic script to run the parallel timing tests as a batch.

Example:
python parallel_timing_batch.py \
    --opsim_file='../data/opsim/baseline_v5.1.1_10yrs.db' \
    --passband_table_dir='../data/passbands/LSST' \
    --threads='1,2,4,8' \
    --samples='10000,20000,30000,40000,50000' \
    --num_simulations=10 \
    --output='parallel_timing_results.npz'
"""

import argparse
import functools
import timeit

import numpy as np
from lightcurvelynx.astro_utils.passbands import PassbandGroup
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc
from lightcurvelynx.math_nodes.ra_dec_sampler import ApproximateMOCSampler
from lightcurvelynx.models.sncosmo_models import SncosmoWrapperModel
from lightcurvelynx.obstable.opsim import OpSim
from lightcurvelynx.simulate import simulate_lightcurves
from lightcurvelynx.utils.extrapolate import LinearDecay


def run_timing_tests(args):
    """Run the timing tests based on the input arguments.

    Parameters
    ----------
    args : `argparse.Namespace`
        The input arguments.
    """
    # Load the OpSim data and passband group.
    print(f"Loading OpSim data from {args.opsim_file}...")
    ops_data = OpSim.from_db(args.opsim_file)
    print(f"Loaded OpSim data with {len(ops_data)} entries.")

    t_min, t_max = ops_data.time_bounds()
    print(f"Time bounds of OpSim data: {t_min} to {t_max}")

    print("Loading passband group 'LSST'...")
    passband_group = PassbandGroup.from_preset(
        preset="LSST",
        table_dir=args.passband_table_dir,
    )
    print(f"Loaded passband group with {len(passband_group)} passbands.")

    # Construct the model.
    print("Constructing the SN model...")
    ra_dec_sampler = ApproximateMOCSampler.from_obstable(ops_data, depth=6)
    source = SncosmoWrapperModel(
        "salt2-h17",
        t0=NumpyRandomFunc("uniform", low=t_min, high=t_max),
        x0=1.0e-6,
        x1=0.5,
        c=0.2,
        ra=ra_dec_sampler.ra,
        dec=ra_dec_sampler.dec,
        redshift=0.01,
        node_label="source",
        time_extrapolation=LinearDecay(decay_width=50.0),
    )

    # Set up the lists of trials to run.
    num_samples_list = [int(x) for x in args.samples.split(",")]
    num_threads_list = [int(x) for x in args.threads.split(",")]
    num_simulations = args.num_simulations
    all_times = np.zeros((len(num_samples_list), len(num_threads_list), num_simulations))
    print("Simulation:")
    print(f"  Samples={num_samples_list}")
    print(f"  Threads={num_threads_list}")
    print(f"  Num simulations={num_simulations}")

    # Loop through each trial and time the simulations for each combination of num_samples and num_threads.
    for sample_idx, num_samples in enumerate(num_samples_list):
        for thread_idx, num_threads in enumerate(num_threads_list):
            # Create a partial function for these parameters.
            simulate_func = functools.partial(
                simulate_lightcurves,
                model=source,
                num_samples=num_samples,
                obstable=ops_data,
                passbands=passband_group,
                num_jobs=num_threads,
                obs_time_window_offset=(-100, 400),
                progress_bar=False,  # Disable progress bar
            )

            # Time each of the simulations and store the results.
            for trial in range(num_simulations):
                t = timeit.timeit(simulate_func, number=1)
                all_times[sample_idx, thread_idx, trial] = t

                # Display each timing result so users can follow along.
                print(f"Time for num_samples={num_samples}, num_threads={num_threads}: {t:.2f} seconds")

            # Output summary timing results per combination of num_samples and num_threads.
            total_time = np.sum(all_times[sample_idx, thread_idx])
            average_time = np.mean(all_times[sample_idx, thread_idx])
            print(
                f"Time (sec) for num_samples={num_samples}, num_threads={num_threads}: "
                f"Total={total_time:.2f}. Average={average_time:.2f}"
            )

    # Save the timing results to a .npz file for later analysis.
    np.savez(
        args.output, all_times=all_times, num_samples_list=num_samples_list, num_threads_list=num_threads_list
    )
    print(f"Saved timing results to {args.output}")


def main():
    """The main function to run the timing tests."""
    parser = argparse.ArgumentParser(
        prog="lightcurvelynx_parallel_timing",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--opsim_file",
        help="Path to the ObsTable file containing the OpSim data.",
        default="../data/opsim/baseline_v5.1.1_10yrs.db",
    )
    parser.add_argument(
        "--passband_table_dir",
        help="Directory containing the passband tables.",
        default="../data/passbands/LSST",
    )
    parser.add_argument(
        "--threads",
        type=str,
        default="1,2,4",
        help="A comma separated list of the number of threads to test.",
    )
    parser.add_argument(
        "--samples",
        type=str,
        default="10000,20000,30000,40000,50000",
        help="A comma separated list of the number of samples to test.",
    )
    parser.add_argument(
        "--output",
        type=str,
        default="parallel_timing_results.npz",
        help="Path to save the timing results as a .npz file.",
    )
    parser.add_argument(
        "--num_simulations",
        type=int,
        default=10,
        help="Number of simulations to run for each combination of threads and samples.",
    )

    # Run the actual program.
    args = parser.parse_args()
    run_timing_tests(args)


if __name__ == "__main__":
    main()



import numpy as np
import pytest
from lightcurvelynx.models.eztaox_models import EzTaoXWrapperModel

# We need to check if eztaox is installed before running the tests that depend on it
try:
    import eztaox.kernels.quasisep as ekq

    has_eztaox = True
except ImportError:
    ekq = None
    has_eztaox = False


@pytest.mark.skipif(not has_eztaox, reason="eztaox is not installed")
def test_eztaox_wrapper_model():
    """
    Test creating a simple EzTaoXWrapperModel and evaluating it at some
    times and wavelengths.
    """
    kernel = ekq.Exp(scale=2.2, sigma=0.02)

    source = EzTaoXWrapperModel(
        kernel,  # The kernel to use
        baseline_mags={"g": 20.0, "r": 19.5, "i": 19.0},
        band_list=["g", "r", "i"],
        log_kernel_param=[1.61, 0.0],
        log_amp_scale=[0.0, 0.0, 0.0],
        zero_mean=True,
        has_lag=False,
        ra=1.0,
        dec=1.0,
        redshift=0.1,
        node_label="source",
    )

    sampled_state = source.sample_parameters(num_samples=1)
    assert "source.eztaox_baseline_mag_g" in sampled_state
    assert "source.eztaox_baseline_mag_r" in sampled_state
    assert "source.eztaox_baseline_mag_i" in sampled_state
    assert "source.eztaox_baseline_mag_z" not in sampled_state
    assert "source.eztaox_log_kernel_param_0" in sampled_state
    assert "source.eztaox_log_kernel_param_1" in sampled_state
    assert "source.eztaox_log_kernel_param_2" not in sampled_state
    assert "source.eztaox_log_amp_scale_0" in sampled_state
    assert "source.eztaox_log_amp_scale_1" in sampled_state
    assert "source.eztaox_log_amp_scale_2" in sampled_state
    assert "source.eztaox_log_amp_scale_3" not in sampled_state

    # Evaluate the model at some times and wavelengths
    times = np.array([0.0, 1.0, 2.0, 3.0])
    filters = np.array(["g", "r", "g", "i"])
    bandfluxes = source.evaluate_bandfluxes(None, times, filters, sampled_state)
    assert bandfluxes.shape == (4,)
    assert np.all(bandfluxes > 0.0)

    # We can evaluate without a given graph state (a new one will be sampled).
    bandfluxes = source.evaluate_bandfluxes(None, times, filters, None)
    assert bandfluxes.shape == (4,)
    assert np.all(bandfluxes > 0.0)

    # We fail if the log_amp_scale length is incorrect
    with pytest.raises(ValueError):
        _ = EzTaoXWrapperModel(
            kernel,  # The kernel to use
            baseline_mags={"g": 20.0, "r": 19.5, "i": 19.0},
            band_list=["g", "r", "i"],
            log_kernel_param=[1.61, 0.0],
            log_amp_scale=[0.0, 0.0],  # Incorrect length
            zero_mean=True,
            has_lag=False,
            ra=1.0,
            dec=1.0,
            redshift=0.1,
            node_label="source",
        )


@pytest.mark.skipif(not has_eztaox, reason="eztaox is not installed")
def test_eztaox_wrapper_model_complex():
    """
    Test creating a simple EzTaoXWrapperModel with lag and non-zero mean, then
    evaluate it at some times and wavelengths
    """
    kernel = ekq.Exp(scale=2.2, sigma=0.02)

    source = EzTaoXWrapperModel(
        kernel,  # The kernel to use
        band_list=["g", "r", "i"],
        log_kernel_param=[1.61, 0.0],
        log_amp_scale=[0.0, 0.0, 0.0],
        zero_mean=False,
        mean_mag=[20.0, 19.5],
        has_lag=True,
        lag=[5.0, 10.0, 15.0],
        ra=1.0,
        dec=1.0,
        redshift=0.1,
        node_label="source",
    )

    sampled_state = source.sample_parameters(num_samples=1)

    # Evaluate the model at some times and wavelengths
    times = np.array([0.0, 1.0, 2.0, 3.0])
    filters = np.array(["g", "r", "g", "i"])
    bandfluxes = source.evaluate_bandfluxes(None, times, filters, sampled_state)
    assert bandfluxes.shape == (4,)
    assert np.all(bandfluxes > 0.0)

    # We fail with incorrect shapes of mean mag or lag.
    with pytest.raises(ValueError):
        _ = EzTaoXWrapperModel(
            kernel,  # The kernel to use
            band_list=["g", "r", "i"],
            log_kernel_param=[1.61, 0.0],
            log_amp_scale=[0.0, 0.0, 0.0],
            zero_mean=False,
            mean_mag=[20.0, 19.5, 18.0],  # Incorrect length
            has_lag=True,
            lag=[5.0, 10.0, 15.0],
            ra=1.0,
            dec=1.0,
            redshift=0.1,
            node_label="source",
        )
    with pytest.raises(ValueError):
        _ = EzTaoXWrapperModel(
            kernel,  # The kernel to use
            band_list=["g", "r", "i"],
            log_kernel_param=[1.61, 0.0],
            log_amp_scale=[0.0, 0.0, 0.0],
            zero_mean=False,
            mean_mag=[20.0, 19.5],
            has_lag=True,
            lag=[5.0, 10.0],  # Incorrect length
            ra=1.0,
            dec=1.0,
            redshift=0.1,
            node_label="source",
        )

    # We fail if we provide both baseline_mags and mean_mag
    with pytest.raises(ValueError):
        _ = EzTaoXWrapperModel(
            kernel,  # The kernel to use
            baseline_mags={"g": 20.0, "r": 19.5, "i": 19.0},
            band_list=["g", "r", "i"],
            log_kernel_param=[1.61, 0.0],
            log_amp_scale=[0.0, 0.0, 0.0],
            zero_mean=False,
            mean_mag=[20.0, 19.5],
            has_lag=False,
            ra=1.0,
            dec=1.0,
            redshift=0.1,
            node_label="source",
        )


@pytest.mark.skipif(not has_eztaox, reason="eztaox is not installed")
def test_eztaox_wrapper_model_seed():
    """
    Test creating a simple EzTaoXWrapperModel with a given seed.
    """
    kernel = ekq.Exp(scale=2.2, sigma=0.02)
    times = np.array([0.0, 1.0, 2.0, 3.0])
    filters = np.array(["g", "r", "g", "i"])

    # Create and evaluate a model.
    source1 = EzTaoXWrapperModel(
        kernel,  # The kernel to use
        baseline_mags={"g": 20.0, "r": 19.5, "i": 19.0},
        band_list=["g", "r", "i"],
        log_kernel_param=[1.61, 0.0],
        log_amp_scale=[0.0, 0.0, 0.0],
        zero_mean=True,
        has_lag=False,
        ra=1.0,
        dec=1.0,
        redshift=0.1,
        seed_param=42,
        node_label="source",
    )
    sampled_state1 = source1.sample_parameters(num_samples=1)
    bandfluxes1 = source1.evaluate_bandfluxes(None, times, filters, sampled_state1)

    # Create a second model with the same seed and evaluate it.
    source2 = EzTaoXWrapperModel(
        kernel,  # The kernel to use
        baseline_mags={"g": 20.0, "r": 19.5, "i": 19.0},
        band_list=["g", "r", "i"],
        log_kernel_param=[1.61, 0.0],
        log_amp_scale=[0.0, 0.0, 0.0],
        zero_mean=True,
        has_lag=False,
        ra=1.0,
        dec=1.0,
        redshift=0.1,
        seed_param=42,
        node_label="source",
    )
    sampled_state2 = source2.sample_parameters(num_samples=1)
    bandfluxes2 = source2.evaluate_bandfluxes(None, times, filters, sampled_state2)

    assert np.allclose(bandfluxes1, bandfluxes2)



import numpy as np
import pytest
from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.models.bagle_models import BagleMultiWrapperModel, BagleWrapperModel


class _FlatBagleModel:
    """A flat dummy bagle model for testing purposes.

    Attributes
    ----------
    scale : float
        A scaling factor for the fluxes.
    filter_mags : list
        A list of magnitudes for each filter.
    """

    def __init__(self, scale, filter_mags, t0=None):
        self.scale = scale
        self.filter_mags = filter_mags

    def get_photometry(self, times, filter_idx):
        """Compute dummy fluxes based on parameters and times.

        Parameters
        ----------
        times : numpy.ndarray
            An array of times.
        filter_idx : int
            The index of the filter.
        """
        return np.ones_like(times) * self.scale * self.filter_mags[filter_idx]


class _GaussianBagleModel:
    """A Gaussian-shaped dummy bagle model for testing purposes.

    Attributes
    ----------
    scale : float
        A scaling factor for the fluxes.
    filter_mags : list
        A list of magnitudes for each filter.
    shift : float
        A constant magnitude shift to apply.
    """

    def __init__(self, scale, filter_mags, t0=None, shift=None, **kwargs):
        self.scale = scale
        self.filter_mags = filter_mags
        self.shift = shift

        # Simulate handling a t0 that is computed based on arguments.
        if t0 is not None:
            self.t0 = t0
        elif "t0_com" in kwargs:
            self.t0 = kwargs["t0_com"] - 2.0
        else:
            self.t0 = 0.0

    def get_photometry(self, times, filter_idx):
        """Compute dummy fluxes based on parameters and times. Simulates
        a Gaussian light curve shape.

        Parameters
        ----------
        times : numpy.ndarray
            An array of times.
        filter_idx : int
            The index of the filter.
        """
        print(f"Computing photometry with t0 = {self.t0}")

        # Since we are computing magnitudes we subtract out the Gaussian shape.
        base = -np.exp(-0.5 * ((times - self.t0) / 50.0) ** 2)
        print(f"Base values: {base}")
        values = base * self.scale + self.filter_mags[filter_idx]
        print(f"Values before shift: {values}")
        if self.shift is not None:
            values += self.shift
        print(f"Values after shift: {values}")
        return values


def test_bagle_wrapper_model() -> None:
    """Test that we can create and query a BagleWrapperModel object."""
    parameter_dict = {
        "scale": 2.0,
        "filter_mags": [20.0, 21.0, 22.0, 23.0, 24.0, 25.0],  # ugrizy
    }
    model = BagleWrapperModel(_FlatBagleModel, parameter_dict, node_label="model")

    # Check that we have the given parameters.
    all_params = model.list_params()
    assert "scale" in all_params
    assert "filter_mags" in all_params
    assert model.parameter_names == ["scale", "filter_mags"]

    # Check that we have the standard physical model parameters.
    assert "ra" in all_params
    assert "dec" in all_params
    assert "t0" in all_params
    assert "redshift" in all_params
    assert "distance" in all_params

    # Test that we can query the model for fluxes.
    graph_state = model.sample_parameters(num_samples=1)
    assert graph_state["model"]["scale"] == 2.0
    assert np.array_equal(
        graph_state["model"]["filter_mags"],
        [20.0, 21.0, 22.0, 23.0, 24.0, 25.0],
    )
    assert graph_state["model"]["ra"] is None
    assert graph_state["model"]["dec"] is None

    query_times = np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.0, 20.0, 21.0])
    query_filters = np.array(["u", "r", "u", "r", "u", "r", "u", "r"])

    fluxes = model.evaluate_bandfluxes(None, query_times, query_filters, graph_state)
    expected_mags = 2.0 * np.array([20.0, 22.0, 20.0, 22.0, 20.0, 22.0, 20.0, 22.0])
    expected_fluxes = mag2flux(expected_mags)
    assert np.allclose(fluxes, expected_fluxes)


def test_bagle_wrapper_model_ra_dec() -> None:
    """Test that we can create and query a BagleWrapperModel object
    with different settings for RA, dec."""
    # No RA/dec given.
    parameter_dict = {
        "scale": 2.0,
        "filter_mags": [20.0, 21.0, 22.0, 23.0, 24.0, 25.0],  # ugrizy
    }
    model1 = BagleWrapperModel(
        _FlatBagleModel,
        parameter_dict,
        node_label="model",
    )
    graph_state1 = model1.sample_parameters(num_samples=1)
    assert graph_state1["model"]["ra"] is None
    assert graph_state1["model"]["dec"] is None

    # Specify RA and dec as kwargs to the wrapper model.
    model2 = BagleWrapperModel(
        _FlatBagleModel,
        parameter_dict,
        ra=21.0,
        dec=-11.0,
        node_label="model",
    )
    graph_state2 = model2.sample_parameters(num_samples=1)
    assert graph_state2["model"]["ra"] == 21.0
    assert graph_state2["model"]["dec"] == -11.0

    # Specify RA and dec as part of the parameter_dict (to the bagle model).
    parameter_dict2 = {
        "scale": 2.0,
        "filter_mags": [20.0, 21.0, 22.0, 23.0, 24.0, 25.0],  # ugrizy
        "raL": 150.0,
        "decL": 2.0,
    }
    model3 = BagleWrapperModel(
        _FlatBagleModel,
        parameter_dict2,
        node_label="model",
    )
    graph_state3 = model3.sample_parameters(num_samples=1)
    assert graph_state3["model"]["ra"] == 150.0
    assert graph_state3["model"]["dec"] == 2.0

    # Using both kwargs and parameter_dict should raise an error.
    with pytest.raises(ValueError):
        _ = BagleWrapperModel(
            _FlatBagleModel,
            parameter_dict2,
            ra=21.0,
            dec=-11.0,
            node_label="model",
        )


def test_bagle_wrapper_model_t0() -> None:
    """Test that we can create and query a BagleWrapperModel object with t0 set."""
    parameter_dict = {
        "scale": 5.0,
        "filter_mags": [20.0, 21.0, 22.0, 23.0, 24.0, 25.0],  # ugrizy
        "t0": 10.0,
    }
    model = BagleWrapperModel(
        _GaussianBagleModel,
        parameter_dict,
        node_label="model",
    )

    # Check that we can query the model for fluxes.
    graph_state = model.sample_parameters(num_samples=1)
    assert graph_state["model"]["t0"] == 10.0

    query_times = np.arange(-5.0, 25.0, 0.1)
    query_filters = np.array(["r"] * len(query_times))
    fluxes = model.evaluate_bandfluxes(None, query_times, query_filters, graph_state)

    # The max flux should be at t0 = 10.0
    max_idx = np.int64(150)  # index of time = 10.0
    assert np.argmax(fluxes) == max_idx

    # We do not allow specifying t0 via kwargs to the wrapper model.
    with pytest.raises(ValueError):
        _ = BagleWrapperModel(
            _GaussianBagleModel,
            parameter_dict,
            t0=10.0,
            node_label="model",
        )

    # We can specify t0 as a computed parameter.
    parameter_dict2 = {
        "scale": 2.0,
        "filter_mags": [20.0, 21.0, 22.0, 23.0, 24.0, 25.0],  # ugrizy
        "t0_com": 12.0,  # t0 will be computed as t0_com - 2.0 = 10.0
    }
    model2 = BagleWrapperModel(
        _GaussianBagleModel,
        parameter_dict2,
        node_label="model",
    )

    # The wrapper model will save the base value of t0 (not the computed one).
    # However the correct translation (computed value) will be used for the fluxes.
    graph_state2 = model2.sample_parameters(num_samples=1)
    assert graph_state2["model"]["t0"] == 12.0  # base value
    fluxes2 = model2.evaluate_bandfluxes(None, query_times, query_filters, graph_state2)
    assert graph_state2["model"]["t0"] == 12.0  # base value
    assert np.argmax(fluxes2) == max_idx


def test_bagle_wrapper_model_filter_idx() -> None:
    """Test that we can create and query a BagleWrapperModel object with filter indices."""
    parameter_dict = {
        "scale": 2.0,
        "filter_mags": [20.0, 21.0, 22.0, 23.0, 24.0, 25.0],  # ugrizy
    }
    custom_filter_idx = {"u": 5, "g": 4, "r": 3, "i": 2, "z": 1, "y": 0}
    model = BagleWrapperModel(_FlatBagleModel, parameter_dict, filter_idx=custom_filter_idx)

    # Test that we can query the model for fluxes.
    graph_state = model.sample_parameters(num_samples=1)
    query_times = np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.0, 20.0, 21.0])
    query_filters = np.array(["u", "r", "u", "r", "u", "r", "u", "r"])

    fluxes = model.evaluate_bandfluxes(None, query_times, query_filters, graph_state)
    expected_mags = 2.0 * np.array([25.0, 23.0, 25.0, 23.0, 25.0, 23.0, 25.0, 23.0])
    expected_fluxes = mag2flux(expected_mags)
    assert np.allclose(fluxes, expected_fluxes)


def test_bagle_multi_wrapper_model() -> None:
    """Test that we can create and query a BagleMultiWrapperModel object."""
    # Create three different bagle models with different parameters (including
    # a different number of parameters per model).
    parameter_dicts = [
        {
            "scale": 2.0,
            "filter_mags": [20.0, 21.0, 22.0, 23.0, 24.0, 25.0],  # ugrizy
            "t0": 0.0,
        },
        {
            "scale": 3.0,
            "filter_mags": [20.0, 20.0, 20.0, 20.0, 20.0, 20.0],  # ugrizy
            "t0": 3.0,
            "raL": 150.0,
            "decL": 2.0,
        },
        {
            "scale": 4.0,
            "filter_mags": [20.0, 20.0, 20.0, 20.0, 20.0, 20.0],  # ugrizy
            "shift": 1.0,
            "t0": 4.0,
        },
    ]
    model_list = [_FlatBagleModel, _GaussianBagleModel, _GaussianBagleModel]

    model = BagleMultiWrapperModel(
        model_list,
        parameter_dicts,
        in_order=True,
        node_label="model",
    )

    # Check that we have the given parameters.
    all_params = model.list_params()
    assert "scale" in all_params
    assert "filter_mags" in all_params
    assert "shift" in all_params

    # Check that we have the standard physical model parameters.
    assert "ra" in all_params
    assert "dec" in all_params
    assert "t0" in all_params
    assert "redshift" in all_params
    assert "distance" in all_params

    # Test that we can query the model for fluxes.
    graph_state = model.sample_parameters(num_samples=3)
    assert np.allclose(graph_state["model"]["scale"], [2.0, 3.0, 4.0])
    assert np.allclose(graph_state["model"]["t0"], [0.0, 3.0, 4.0])
    assert np.array_equal(
        graph_state["model"]["filter_mags"][0],
        [20.0, 21.0, 22.0, 23.0, 24.0, 25.0],
    )
    assert np.allclose(graph_state["model"]["filter_mags"][1], 20.0)
    assert np.allclose(graph_state["model"]["filter_mags"][2], 20.0)
    assert np.array_equal(graph_state["model"]["ra"], [None, 150.0, None])
    assert np.array_equal(graph_state["model"]["dec"], [None, 2.0, None])

    query_times = np.array([0.0, 0.5, 1.0, 2.0, 3.0, 4.0, 20.0, 21.0])
    query_filters = np.array(["u", "r", "u", "r", "u", "r", "u", "r"])

    fluxes = model.evaluate_bandfluxes(None, query_times, query_filters, graph_state)
    print(fluxes)
    assert fluxes.shape == (3, len(query_times))

    # Check the flat model.
    mag_0 = 2.0 * np.array([20.0, 22.0, 20.0, 22.0, 20.0, 22.0, 20.0, 22.0])
    assert np.allclose(fluxes[0], mag2flux(mag_0))

    # Check the Gaussian models have the correct peak.
    assert np.argmax(fluxes[1]) == 4  # max at t0 = 3.0
    mag_1 = -np.exp(-0.5 * ((query_times - 3.0) / 50.0) ** 2) * 3.0 + 20.0
    assert np.allclose(fluxes[1], mag2flux(mag_1))

    assert np.argmax(fluxes[2]) == 5  # max at t0 = 4.0
    mag_2 = -np.exp(-0.5 * ((query_times - 4.0) / 50.0) ** 2) * 4.0 + 21.0
    assert np.allclose(fluxes[2], mag2flux(mag_2))



"""Utility functions for post processing the results data by adding statistics
columns and filtering on those columns."""

import warnings

import numpy as np
import numpy.ma as ma
import pandas as pd
from nested_pandas import NestedFrame

from lightcurvelynx.astro_utils.mag_flux import flux2mag
from lightcurvelynx.obstable.obs_table import ObsTable


def concat_results(results_list):
    """Concatenate a list of results into a single NestedFrame,
    updating the ID column to be unique across all results.

    Parameters
    ----------
    results_list : list of nested_pandas.NestedFrame
        The list of DataFrames to concatenate.

    Returns
    -------
    nested_pandas.NestedFrame
        The concatenated DataFrame.
    """
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore", category=FutureWarning, message=".*DataFrame concatenation with empty.*"
        )

        result = pd.concat(results_list, ignore_index=True)

        # We need to update the ID column to be unique across all results.
        if "id" in result.columns:
            result["id"] = np.arange(len(result))
        return result


def results_drop_empty(results):
    """Drop empty lightcurves from the results DataFrame.

    Parameters
    ----------
    results : nested_pandas.NestedFrame
        The DataFrame containing lightcurve data.

    Returns
    -------
    nested_pandas.NestedFrame
        The DataFrame with empty lightcurves removed.
    """
    return results.dropna(subset=["lightcurve"])


def results_append_param_as_col(results, param_name):
    """Append a simulation parameter as a new column to the results DataFrame.

    Parameters
    ----------
    results : nested_pandas.NestedFrame
        The DataFrame containing lightcurve data. This is modified in place.
    param_name : str
        The name of the parameter to append in the form <node_label>.<param_name>.

    Returns
    -------
    nested_pandas.NestedFrame
        The DataFrame with the new parameter column added.
    """
    # Remove the dot from the parameter name to create a valid column name.
    new_colname = param_name.replace(".", "_")
    if new_colname in results.columns:
        warnings.warn(f"Parameter {new_colname} already exists in results. Overwriting.")

    # Because the parameters are stored as a list of dictionaries, we need to loop through
    # each row and extract the parameter value.
    values = [results["params"].iloc[i][param_name] for i in range(len(results))]
    results[new_colname] = values

    return results


def results_append_obstable_data(results, column_name, obstables):
    """Append the ObsTable entries for each observation as a new column in the
    lightcurves nested DataFrame.

    Parameters
    ----------
    results : nested_pandas.NestedFrame
        The DataFrame containing lightcurve data. This is modified in place.
    column_name : str
        The name of the column to append from the ObsTable entries.
    obstables : ObsTable or list of ObsTable
        The ObsTable(s) containing the data to append. These should be in the
        same order as where used in the simulation.

    Returns
    -------
    nested_pandas.NestedFrame
        The DataFrame with the new parameter column added.
    """
    if isinstance(obstables, ObsTable):
        obstables = [obstables]

    obs_idx = results["lightcurve.obs_idx"]
    survey_idx = results["lightcurve.survey_idx"]
    unique_survey_idx = np.unique(survey_idx)

    new_col = np.full(len(obs_idx), np.nan)
    for s_idx in unique_survey_idx:
        survey_mask = survey_idx == s_idx
        if column_name in obstables[s_idx].columns:
            new_col[survey_mask] = obstables[s_idx][column_name].iloc[obs_idx[survey_mask]]
    results[f"lightcurve.{column_name}"] = new_col

    return results


def lightcurve_compute_snr(flux, fluxerr):
    """Compute the signal-to-noise ratio (SNR) for given flux and flux error arrays.

    Parameters
    ----------
    flux : array-like
        The flux values.
    fluxerr : array-like
        The flux error values.

    Returns
    -------
    result : np.ndarray
        The SNR values, with None for invalid entries (e.g., zero or negative flux error).
    """
    flux = np.asarray(flux)
    fluxerr = np.asarray(fluxerr)
    valid_mask = (flux > 0) & (fluxerr > 0)

    result = ma.masked_all(flux.shape)
    result[valid_mask] = flux[valid_mask] / fluxerr[valid_mask]
    return result


def lightcurve_compute_mag(flux, fluxerr):
    """Compute the AB magnitude and magnitude error for given flux and flux error arrays.

    Parameters
    ----------
    flux : array-like
        The flux values.
    fluxerr : array-like
        The flux error values.

    Returns
    -------
    tuple of np.ndarray
        The magnitude and magnitude error values, with None for invalid entries (e.g., non-positive flux).
    """
    flux = np.asarray(flux)
    fluxerr = np.asarray(fluxerr)
    valid_mask = (flux > 0) & (fluxerr > 0)

    mag = ma.masked_all(flux.shape)
    mag[valid_mask] = flux2mag(flux[valid_mask])

    magerr = ma.masked_all(flux.shape)
    magerr[valid_mask] = (2.5 / np.log(10)) * (fluxerr[valid_mask] / flux[valid_mask])

    return mag, magerr


def augment_single_lightcurve(results, *, min_snr=0.0, t0=None):
    """Add columns to a single lightcurve DataFrame with additional information
    about the light curve, including:
    - SNR = flux / fluxerr
    - detection flag (True if SNR >= min_snr, False otherwise)
    - AB magnitude
    - AB magnitude error = (2.5 / ln(10)) * (fluxerr / flux)
    - relative time = mjd - t0 (if t0 is provided)
    None is used for invalid entries, e.g. negative flux or zero flux error.

    Parameters
    ----------
    results : pandas.DataFrame
        The DataFrame containing lightcurve data. Modified in place.
    min_snr : float, optional
        Minimum SNR required to mark an entry as a detection. Default is 0.0.
    t0 : float or None, optional
        Reference time for the lightcurve.

    Returns
    -------
    results : pandas.DataFrame
        The modified DataFrame (to enable chaining).
    """
    if "flux" not in results.columns or "fluxerr" not in results.columns:
        raise ValueError("flux and fluxerr must be present in the light curve DataFrame.")
    flux = results["flux"]
    fluxerr = results["fluxerr"]

    snr = lightcurve_compute_snr(flux, fluxerr)
    results["snr"] = snr
    results["detection"] = np.where(snr.mask, False, snr >= min_snr)

    mag, magerr = lightcurve_compute_mag(flux, fluxerr)
    results["mag"] = mag
    results["magerr"] = magerr

    if t0 is not None and "mjd" in results.columns:
        results["time_rel"] = results["mjd"] - t0

    return results


def results_augment_lightcurves(results, *, min_snr=0.0):
    """Add columns to the results DataFrame with additional information
    about each light curve, including:
    - SNR = flux / fluxerr
    - detection flag (True if SNR >= min_snr, False otherwise)
    - AB magnitude
    - AB magnitude error = (2.5 / ln(10)) * (fluxerr / flux)
    - relative time = mjd - t0 (if t0 in the results table)
    None is used for invalid entries, e.g. negative flux or zero flux error.

    The input data frame can either be a single light curve (pandas.DataFrame)
    with columns "flux" and "fluxerr", or a NestedFrame (nested_pandas.NestedFrame)
    with a nested DataFrame column "lightcurve" that contains the "flux" and
    "fluxerr" columns.

    Parameters
    ----------
    results : pandas.DataFrame or nested_pandas.NestedFrame
        The DataFrame containing lightcurve data. Modified in place.
    min_snr : float, optional
        Minimum SNR required to mark an entry as a detection. Default is 0.0.

    Returns
    -------
    results : pandas.DataFrame or nested_pandas.NestedFrame
        The modified DataFrame (to enable chaining).
    """
    if not isinstance(results, NestedFrame) or "lightcurve" not in results.columns:
        raise ValueError("results must be a NestedFrame with a 'lightcurve' column.")
    if (
        "flux" not in results["lightcurve"].nest.columns
        or "fluxerr" not in results["lightcurve"].nest.columns
    ):
        raise ValueError("lightcurve.flux and lightcurve.fluxerr must be present in the DataFrame.")
    flux = results["lightcurve.flux"]
    fluxerr = results["lightcurve.fluxerr"]

    # Compute SNR and detection flag.
    snr = lightcurve_compute_snr(flux, fluxerr)
    results["lightcurve.snr"] = snr
    results["lightcurve.detection"] = np.where(snr.mask, False, snr >= min_snr)

    # Compute magnitude and magnitude error.
    mag, magerr = lightcurve_compute_mag(flux, fluxerr)
    results["lightcurve.mag"] = mag
    results["lightcurve.magerr"] = magerr

    # If t0 is provided as a column in results, compute relative time.
    if "t0" in results.columns and np.all(results["t0"]) and results["t0"].notna().all():
        if "mjd" not in results["lightcurve"].nest.columns:
            raise ValueError("lightcurve.mjd must be present in the DataFrame.")

        # Get the index for the t0 entry for each lightcurve MJD and use that
        # to subtract out the reference t0.
        t0 = np.asanyarray(results["t0"])
        t0_idx = np.array(results["lightcurve"]["mjd"].index)
        results["lightcurve.time_rel"] = results["lightcurve.mjd"] - t0[t0_idx]

    return results


def results_use_full_filter_names(results, passbands):
    """Modifies the 'filter' column in the results DataFrame to include
    the survey name as a prefix, e.g. 'LSST_g'.

    Parameters
    ----------
    results : pandas.DataFrame or nested_pandas.NestedFrame
        The DataFrame containing lightcurve data. Modified in place.
    passbands : list of PassbandGroup
        The list of PassbandGroups used in the simulation, in the same order
        as in the simulation.

    Returns
    -------
    results : pandas.DataFrame or nested_pandas.NestedFrame
        The modified DataFrame (to enable chaining).
    """
    if not isinstance(results, NestedFrame) or "lightcurve" not in results.columns:
        raise ValueError("results must be a NestedFrame with a 'lightcurve' column.")

    if "filter" not in results["lightcurve"].nest.columns:
        raise ValueError("lightcurve.flux and lightcurve.fluxerr must be present in the DataFrame.")

    if "survey_idx" in results["lightcurve"].nest.columns:
        survey_idx = results["lightcurve.survey_idx"].values
    else:
        survey_idx = np.zeros(len(results), dtype=int)

    # Go through every pair of survey index and filter name and replace
    # the filter name with the full name from the passband group.
    filter_names = results["lightcurve.filter"].values.copy()
    for s_idx in np.unique(survey_idx):
        for fil in np.unique(filter_names[survey_idx == s_idx]):
            mask = (survey_idx == s_idx) & (filter_names == fil)
            full_name = passbands[s_idx][fil].full_name
            filter_names[mask] = full_name
    results["lightcurve.filter"] = filter_names

    return results



import numpy as np
import pandas as pd
import pytest
from lightcurvelynx.obstable.spectrograph_table import SpectrographObsTable


def test_create_spectrograph_obstable():
    """Test initializing SpectrographObsTable."""
    data = {
        "ra": [0.0, 10.0, 0.0, 0.0, 10.0, 0.0],
        "dec": [0.0, -10.0, 0.0, 0.0, -10.0, 0.0],
        "time": 53370.5 + np.array([-5.0, -2.0, 0.0, 5.0, 7.0, 10.0]),
    }
    survey_data_table = pd.DataFrame(data)
    survey_data = SpectrographObsTable(table=survey_data_table)

    assert np.all(survey_data["ra"] == data["ra"])
    assert np.all(survey_data["dec"] == data["dec"])
    assert np.all(survey_data["time"] == data["time"])

    # We fill in fake zp and filter columns. These are not used for anything.
    assert np.all(survey_data["filter"] == "spectra")
    assert np.all(survey_data["zp"] == 0.05)

    # We have all the attributes set at their default values.
    assert survey_data.survey_values["radius"] == pytest.approx(10.0 / 3600.0)



"""A class to hold SED information."""

import numpy as np
from astropy import units as u
from citation_compass import cite_function

from lightcurvelynx.utils.io_utils import read_numpy_data, write_numpy_data


class SED:
    """A class to hold SED information.

    Attributes
    ----------
    wavelengths : np.ndarray
        The wavelength values of the SED.
    fluxes : np.ndarray
        The flux values of the SED.
    """

    def __init__(self, wavelengths, fluxes, **kwargs):
        self.wavelengths = np.array(wavelengths)
        self.fluxes = np.array(fluxes)

        if len(wavelengths) < 2:
            raise ValueError("SED must have at least two wavelength and flux values.")
        if len(wavelengths) != len(fluxes):
            raise ValueError("Wavelengths and fluxes must have the same length.")
        if not np.all(np.diff(wavelengths) > 0):
            raise ValueError("Wavelengths are not in sorted order.")

    @classmethod
    def from_file(cls, sed_file, **kwargs):
        """Load a static SED from a file containing a two column array where the
        first column is wavelength (in angstroms) and the second column is flux (in nJy).

        Parameters
        ----------
        sed_file : str or Path
            The path to the SED file to load.
        **kwargs : dict
            Additional keyword arguments to pass to the SED constructor.

        Returns
        -------
        SED
            An instance of SED with the loaded SED data.
        """
        # Load the SED data from the file (automatically detected format)
        sed_data = read_numpy_data(sed_file)
        if sed_data.ndim != 2 or sed_data.shape[1] != 2:
            raise ValueError(f"SED data from {sed_file} must be a two column array.")

        return cls(sed_data[:, 0], sed_data[:, 1], **kwargs)

    @classmethod
    @cite_function
    def from_synphot(cls, sp_model, waves=None, **kwargs):
        """Generate the spectrum from a given synphot model.

        References
        ----------
        synphot (ascl:1811.001)

        Parameters
        ----------
        sp_model : synphot.SourceSpectrum
            The synphot model to generate the spectrum from.
        waves : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms) at which to sample the SED.
            If None, the SED will be sampled at the wavelengths defined in the synphot model.
        **kwargs : dict
            Additional keyword arguments to pass to the SED constructor.

        Returns
        -------
        SED
            An instance of SED with the generated SED data.
        """
        try:
            from synphot import units
        except ImportError as err:  # pragma: no cover
            raise ImportError(
                "synphot package is not installed be default. To use the synphot models, please "
                "install it. For example, you can install it with `pip install synphot`."
            ) from err

        if sp_model.z > 0.0:
            raise ValueError(
                "The synphot model must be defined at the rest frame (z=0.0). "
                f"Current redshift is {sp_model.z}."
            )

        if waves is None:
            waves = np.array(sp_model.waveset * u.angstrom)

        # Extract the SED data from the synphot model. Synphot models return flux in units
        # of PHOTLAM (photons s^-1 cm^-2 A^-1), so we convert to nJy.
        photlam_flux = sp_model(waves, flux_unit=units.PHOTLAM)
        sed_data = np.array(units.convert_flux(waves, photlam_flux, "nJy"))
        return cls(waves, sed_data, **kwargs)

    def minwave(self):
        """Get the minimum wavelength of the SED.

        Returns
        -------
        minwave : float
            The minimum wavelength of the SED (in angstroms).
        """
        return self.wavelengths[0]

    def maxwave(self):
        """Get the maximum wavelength of the SED.

        Returns
        -------
        maxwave : float
            The maximum wavelength of the SED (in angstroms).
        """
        return self.wavelengths[-1]

    def evaluate(self, wavelengths):
        """Evaluate the SED at the given wavelengths by interpolating the SED data.

        Parameters
        ----------
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).

        Returns
        -------
        flux_density : numpy.ndarray
            A length N matrix of observer frame SED values (in nJy).
        """
        sed_fluxes = np.interp(
            wavelengths,
            self.wavelengths,
            self.fluxes,
            left=0.0,
            right=0.0,
        )
        return sed_fluxes

    def to_file(self, sed_file):
        """Save the SED to a file as a two column array where the first column is wavelength
        (in angstroms) and the second column is flux (in nJy).

        Parameters
        ----------
        sed_file : str or Path
            The path to the SED file to save.
        """
        sed_data = np.column_stack((self.wavelengths, self.fluxes))
        write_numpy_data(sed_file, sed_data)

    def plot(self, *, ax=None, **kwargs):
        """Plot the SED.

        Parameters
        ----------
        ax : matplotlib.axes.Axes, optional
            The axes to plot on. If None, a new figure and axes will be created.
        **kwargs : dict
            Additional keyword arguments to pass to the plot function.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes with the SED plot.
        """
        import matplotlib.pyplot as plt

        if ax is None:
            _, ax = plt.subplots()  # pragma: no cover

        ax.plot(self.wavelengths, self.fluxes, **kwargs)
        ax.set_xlabel("Wavelength (Angstroms)")
        ax.set_ylabel("Flux (nJy)")
        ax.set_title("Spectral Energy Distribution (SED)")
        return ax



import numpy as np
import pytest
from astropy.cosmology import WMAP9, Planck18
from lightcurvelynx.astro_utils.redshift import (
    RedshiftDistFunc,
    obs_to_rest_times_waves,
    redshift_to_distance,
    rest_to_obs_flux,
)
from lightcurvelynx.models.basic_models import StepModel


def test_obs_to_rest_times_waves() -> None:
    """Test that we correctly convert observer frame times and wavelengths to rest frame."""
    observer_times = np.array([10.0, 20.0, 30.0])
    observer_waves = np.array([4000.0, 5000.0, 6000.0])
    t0 = 5.0

    # Test with redshift = 0. This should be a no-op.
    rest_times, rest_waves = obs_to_rest_times_waves(observer_times, observer_waves, 0.0, t0)
    assert np.all(rest_times == observer_times)
    assert np.all(rest_waves == observer_waves)

    # Test with redshift = 1. Times should decrease (relative to t0) and
    # wavelengths should halve.
    rest_times, rest_waves = obs_to_rest_times_waves(observer_times, observer_waves, 1.0, t0)
    expected_rest_times = np.array([7.5, 12.5, 17.5])
    expected_rest_waves = np.array([2000.0, 2500.0, 3000.0])
    assert np.allclose(rest_times, expected_rest_times)
    assert np.allclose(rest_waves, expected_rest_waves)

    # Check that we fail if redshift is negative.
    with pytest.raises(ValueError):
        obs_to_rest_times_waves(observer_times, observer_waves, -0.5, t0)


def test_rest_to_obs_flux() -> None:
    """Test that we correctly convert rest frame flux densities to observer frame."""
    rest_flux = np.array([10.0, 20.0, 30.0])
    redshift = 0.8
    obs_flux = rest_to_obs_flux(rest_flux, redshift)
    expected_obs_flux = rest_flux * (1 + redshift)
    assert np.allclose(obs_flux, expected_obs_flux)

    # Check that we fail if redshift is negative.
    with pytest.raises(ValueError):
        rest_to_obs_flux(rest_flux, -0.5)


def test_redshifted_flux_densities() -> None:
    """Test that we correctly calculate redshifted values."""
    times = np.linspace(0, 100, 1000)
    wavelengths = np.array([100.0, 200.0, 300.0])
    t0 = 10.0
    t1 = 30.0
    brightness = 50.0

    for redshift in [0.0, 0.5, 2.0, 3.0, 30.0]:
        model_redshift = StepModel(brightness=brightness, t0=t0, t1=t1, redshift=redshift)
        values_redshift = model_redshift.evaluate_sed(times, wavelengths)

        for i, time in enumerate(times):
            if t0 <= time and time <= (t1 - t0) * (1 + redshift) + t0:
                # Note that the multiplication by (1+z) is due to the fact we are working in f_nu
                # units, instead of f_lambda units and may be unintuitive for users who are used to
                # working in f_lambda units. This factor can be derived by equaling the integrated
                # flux in f_nu unit before and after redshift is applied.
                assert np.all(values_redshift[i] == brightness * (1 + redshift))
            else:
                assert np.all(values_redshift[i] == 0.0)


def test_redshift_to_distance():
    """Test that we can convert the redshift to a distance using a given cosmology."""
    wmap9_val = redshift_to_distance(1100, cosmology=WMAP9)
    planck18_val = redshift_to_distance(1100, cosmology=Planck18)

    assert abs(planck18_val - wmap9_val) > 1000.0
    assert 13.0 * 1e12 < wmap9_val < 16.0 * 1e12
    assert 13.0 * 1e12 < planck18_val < 16.0 * 1e12

    # We fail with invalid redshift or no cosmology.
    with pytest.raises(ValueError):
        redshift_to_distance(-0.1, cosmology=WMAP9)
    with pytest.raises(ValueError):
        redshift_to_distance(1.0, cosmology=None)


def test_redshift_dist_func_node():
    """Test the RedshiftDistFunc node."""
    node = RedshiftDistFunc(redshift=1100, cosmology=Planck18)
    state = node.sample_parameters()
    assert 13.0 * 1e12 < node.get_param(state, "function_node_result") < 16.0 * 1e12

    # Test that we can generate multiple samples.
    state = node.sample_parameters(num_samples=10)
    assert np.all(node.get_param(state, "function_node_result") > 13.0 * 1e12)
    assert np.all(node.get_param(state, "function_node_result") < 16.0 * 1e12)



import numpy as np
import pandas as pd
import pytest
from lightcurvelynx.obstable.lsst_obstable import LSSTObsTable


def test_create_lsst_obstable():
    """Create a minimal LSSTObsTable object and perform basic queries."""
    values = {
        "expMidptMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "zp": np.ones(5),
    }
    pdf = pd.DataFrame(values)

    ops_data = LSSTObsTable(pdf)
    assert len(ops_data) == 5
    assert len(ops_data.columns) == 4

    # We have all the attributes set at their default values.
    assert ops_data.survey_values["dark_current"] == 0.022
    assert ops_data.survey_values["gain"] == 1.595
    assert ops_data.survey_values["pixel_scale"] == 0.2
    assert ops_data.survey_values["radius"] == 1.75
    assert ops_data.survey_values["read_noise"] == 5.82
    assert ops_data.survey_values["survey_name"] == "LSST"

    # Check that we can extract the time bounds.
    t_min, t_max = ops_data.time_bounds()
    assert t_min == 0.0
    assert t_max == 4.0

    # We can query which columns the LSSTObsTable has by new or old name.
    assert "ra" in ops_data
    assert "dec" in ops_data
    assert "time" in ops_data
    assert "expMidptMJD" in ops_data

    # We can access columns directly as though it was a table.
    assert np.allclose(ops_data["ra"], values["ra"])
    assert np.allclose(ops_data["dec"], values["dec"])
    assert np.allclose(ops_data["time"], values["expMidptMJD"])
    assert np.allclose(ops_data["expMidptMJD"], values["expMidptMJD"])

    # Without a filters column we cannot access the filters.
    assert len(ops_data.filters) == 0


def test_create_lsst_obstable_override_fail():
    """Test that we fail if we do not have the information needed to create the zeropoints."""
    values = {
        "expMidptMJD": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }

    with pytest.raises(ValueError):
        _ = LSSTObsTable(values, ext_coeff=None)


def _make_fake_data(times):
    """Create 9 fake ccd pointings at each timestep using
    a 3x3 grid centered (like dp1) with grid steps of 0.22 deg.

    Parameters
    ----------
    times : `numpy.ndarray`
        The times of the observations (in MJD).
    """

    # Create a tile of pointings around a random center at each time
    # where the center is chosen from a 1 degree by 1 degree box.
    data = {
        "expMidptMJD": [],
        "ra": [],
        "dec": [],
        "band": [],
    }
    for t in times:
        pointing_ra = np.random.uniform(150.0, 151.0)
        pointing_dec = np.random.uniform(-21.0, -20.0)
        band = np.random.choice(["g", "r", "i"])

        for dra in [-0.22, 0.0, 0.22]:
            for ddec in [-0.22, 0.0, 0.22]:
                data["expMidptMJD"].append(t)
                data["ra"].append(pointing_ra + dra)
                data["dec"].append(pointing_dec + ddec)
                data["band"].append(band)

    # Add the other columns with random values.
    num_samples = len(data["expMidptMJD"])
    data["ccdVisitId"] = np.arange(num_samples)
    data["expTime"] = np.full(num_samples, 30.0)  # seconds
    data["magLim"] = np.random.normal(24.0, 0.5, num_samples)  # mag
    data["seeing"] = np.random.normal(1.0, 0.2, num_samples)  # arcseconds
    data["skyBg"] = np.random.normal(1750, 10.0, num_samples)  # adu
    data["skyNoise"] = np.random.normal(43.0, 1.0, num_samples)  # adu
    data["skyRotation"] = np.zeros(num_samples)  # degrees
    data["pixelScale"] = np.full(num_samples, 0.2)  # arcsec/pixel
    data["xSize"] = np.full(num_samples, 4000)  # pixels
    data["ySize"] = np.full(num_samples, 4000)  # pixels
    data["zeroPoint"] = np.random.normal(31.0, 0.2, num_samples)  # mag

    return pd.DataFrame(data)


def test_lsst_obstable_from_ccdvisit():
    """Test that we can create an LSSTObsTable from a CCD visit database."""
    times = 60623.25 + np.arange(20) * 0.1  # 20 time steps (15 minutes apart)
    ccd_visit_table = _make_fake_data(times)
    obs_table = LSSTObsTable.from_ccdvisit_table(ccd_visit_table)
    assert len(obs_table) == 180  # Number of rows in the test ccdvisit table

    # Check that we have the expected columns.
    assert "time" in obs_table
    assert "ra" in obs_table
    assert "dec" in obs_table
    assert "filter" in obs_table
    assert "zp" in obs_table
    assert "maglim" in obs_table
    assert "seeing" in obs_table
    assert "rotation" in obs_table
    assert "radius" in obs_table

    # Check that we have the expected survey radius values.
    assert np.all(obs_table["radius"] >= 0.15)
    assert np.all(obs_table["radius"] <= 0.16)
    assert np.all(obs_table["pixel_scale"] >= 0.19)
    assert np.all(obs_table["pixel_scale"] <= 0.21)

    # No footprint is created by default.
    assert obs_table._detector_footprint is None

    # If we create a footprint, it is set correctly.
    obs_table_with_footprint = LSSTObsTable.from_ccdvisit_table(
        ccd_visit_table,
        make_detector_footprint=True,
    )
    assert obs_table_with_footprint._detector_footprint is not None


def test_lsst_obstable_from_sv_visits():
    """Test that we can read an LSSTObsTable from the SV visits table."""
    values = {
        "exp_midpt_mjd": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "fieldRA": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "fieldDec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "sky_bg_median": np.ones(5),
        "zero_point_median": 25.0 * np.ones(5),
    }
    pdf = pd.DataFrame(values)

    ops_data = LSSTObsTable.from_sv_visits_table(pdf)
    assert len(ops_data) == 5
    assert ops_data.radius == pytest.approx(1.75)



import gzip
import logging
import os
import sqlite3
import sys
from pathlib import Path

import numpy as np
import pandas as pd
from astropy.table import Table


class SquashOutput:
    """Context manager to temporarily squash all output to stdout and stderr.

    Optionally, stdout can be redirected to a logger instead of being suppressed.

    Parameters
    ----------
    stdout_to_log : bool, optional
        If True, stdout and stderr will be redirected to the logger instead of
        being suppressed.
        Default is False.
    logger : logging.Logger, optional
        The logger to which to redirect the output if stdout_to_log is True.
        If None, the root logger will be used. Default is None.
    log_level : int, optional
        The logging level to use when redirecting stdout and stderr to the logger.
        Default is logging.DEBUG.

    Example
    -------
    with SquashOutput():
        # Code that produces unwanted output
        ...

    with SquashOutput(stdout_to_log=True):
        # Printed output is sent to logger.debug
        ...
    """

    class _LoggerWriter:
        """File-like wrapper that forwards writes to a logger."""

        def __init__(self, logger, level=logging.DEBUG):
            self._logger = logger
            self._level = level

        def write(self, message):
            for line in message.rstrip().splitlines():
                if line:
                    self._logger.log(self._level, line)
            return len(message)

        def flush(self):
            return None

    def __init__(self, stdout_to_log=False, logger=None, log_level=logging.DEBUG):
        self._stdout_to_log = stdout_to_log
        self._log_level = log_level
        self._logger = logger if logger is not None else logging.getLogger()
        self._original_stdout = None
        self._original_stderr = None
        self._null_file = None
        self._stdout_logger = None

    def __enter__(self):
        # Save the original stdout and stderr streams.
        self._original_stdout = sys.stdout
        self._original_stderr = sys.stderr

        # Redirect output streams.
        if self._stdout_to_log:
            # Create a log writter if needed, and redirect stdout and stderr to it.
            if self._stdout_logger is None:
                self._stdout_logger = SquashOutput._LoggerWriter(
                    self._logger,
                    level=self._log_level,
                )
            sys.stdout = self._stdout_logger
            sys.stderr = self._stdout_logger
        else:
            # Open the null file if we haven't already, and redirect stdout and stderr to it.
            if self._null_file is None:
                self._null_file = open(os.devnull, "w")  # noqa: SIM115
            sys.stdout = self._null_file
            sys.stderr = self._null_file

    def __exit__(self, exc_type, exc_value, traceback):
        # Restore the original stdout and stderr streams.
        sys.stdout = self._original_stdout
        sys.stderr = self._original_stderr

        # Close the null file (if we opened one).
        if self._null_file is not None:
            self._null_file.close()  # noqa: SIM115
            self._null_file = None


class SquashLogging:
    """A context manager to temporarily squash all logging (below a certain level)
    to a logger.

    Parameters
    ----------
    logger : logging.Logger
        The logger to which to redirect the logging output.
    level : int, optional
        The logging level below which to suppress logging output. Default: logging.ERROR

    Example
    -------
    with SquashLogging(logger=logging.getLogger(), level=logging.ERROR):
        # Code that produces unwanted logging at any level below logging.ERROR
        ...
    """

    def __init__(self, logger, level=logging.ERROR):
        if logger is None:
            raise ValueError("Logger must be provided to SquashLogging.")
        self._logger = logger
        self._old_level = None
        self._new_level = level

    def __enter__(self):
        # Save the original logging level and set the new level.
        self._old_level = self._logger.level
        self._logger.setLevel(self._new_level)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        # Restore the original logging level.
        self._logger.setLevel(self._old_level)


def write_results_as_hats(base_catalog_path, results, *, catalog_name=None, overwrite=False):
    """Write results to a HATS catalog.

    Parameters
    ----------
    base_catalog_path : str or Path
        The base path to the output hats directory.
    results : nested_pandas.NestedFrame
        The results to write, as a NestedFrame where each row is a sample.
    catalog_name : str, optional
        The name of the catalog to write. If None, the name will be derived from the
        base_catalog_path. Default: None
    overwrite : bool
        Whether to overwrite the output directory if it already exists.
        Default: False
    """
    base_catalog_path = Path(base_catalog_path)
    logging.debug(f"Writing results as HATS Catalog to {base_catalog_path}")

    # See if the (optional) LSDB package is installed.
    try:
        from lsdb import from_dataframe
    except ImportError as err:  # pragma: no cover
        raise ImportError(
            "The lsdb package is required to write results as HATS files. "
            "Please install it via 'pip install lsdb'."
        ) from err

    # Convert the results into an LSDB Catalog and output that. We just generate and output
    # the basic catalog (no margins or other extras).
    catalog = from_dataframe(results, ra_column="ra", dec_column="dec")
    catalog.write_catalog(
        base_catalog_path,
        catalog_name=catalog_name,
        as_collection=False,
        overwrite=overwrite,
    )


def read_numpy_data(file_path):
    """Read in a numpy array from different formats depending on the file extension.
    Automatically detects and handles files in .npy, .npz, .csv, .ecsv, and .txt
    formats.

    Parameters
    ----------
    file_path : str
        The path to the file to read.

    Returns
    -------
    data : numpy.ndarray
        The data read from the file.
    """
    file_path = Path(file_path)
    if not file_path.is_file():
        raise FileNotFoundError(f"File {file_path} not found.")

    # Load the data according to the format.
    if file_path.suffix == ".npy":
        data = np.load(file_path)
    elif file_path.suffix == ".npz":
        # For npz files, extract the first array
        data = np.load(file_path)["arr_0"]
    elif file_path.suffix in [".csv", ".ecsv"]:
        data = np.loadtxt(file_path, delimiter=",", comments="#")
    elif file_path.suffix in [".txt"]:
        data = np.loadtxt(file_path, comments="#")
    else:
        raise ValueError(f"Unsupported file format: {file_path.suffix}.")

    return data


def write_numpy_data(file_path, data):
    """Write a numpy array to a file in a format determined by the file extension.
    Automatically detects and handles files in .npy, .npz, .csv, .ecsv, and .txt
    formats.

    Parameters
    ----------
    file_path : str
        The path to the file to write.
    data : numpy.ndarray
        The data to write to the file.
    """
    file_path = Path(file_path)
    if file_path.suffix == ".npy":
        np.save(file_path, data)
    elif file_path.suffix == ".npz":
        np.savez_compressed(file_path, arr_0=data)
    elif file_path.suffix in [".csv", ".ecsv"]:
        np.savetxt(file_path, data, delimiter=",")
    elif file_path.suffix in [".txt", ".dat"]:
        np.savetxt(file_path, data)
    else:
        raise ValueError(f"Unsupported file format: {file_path.suffix}.")


def read_grid_data(input_file, format="ascii", validate=False):
    """Read 2-d grid data from a text, csv, ecsv, or fits file.

    Each line is of the form 'x0 x1 value' where x0 and x1 are the grid
    coordinates and value is the grid value. The rows should be sorted by
    increasing x0 and, within an x0 value, increasing x1.

    Parameters
    ----------
    input_file : str or file-like object
        The input data file.
    format : str
        The file format. Should be one of the formats supported by
        astropy Tables such as 'ascii', 'ascii.ecsv', or 'fits'.
        Default: 'ascii'
    validate : bool
        Perform additional validation on the input data.
        Default: False

    Returns
    -------
    x0 : numpy.ndarray
        A 1-d array with the values along the x-axis of the grid.
    x1 : numpy.ndarray
        A 1-d array with the values along the y-axis of the grid.
    values : numpy.ndarray
        A 2-d array with the values at each point in the grid with
        shape (len(x0), len(x1)).

    Raises
    ------
    ValueError if any data validation fails.
    """
    logging.debug(f"Loading file {input_file} (format={format})")
    if not Path(input_file).is_file():
        raise FileNotFoundError(f"File {input_file} not found.")

    data = Table.read(input_file, format=format, comment=r"\s*#")
    if len(data.colnames) != 3:
        raise ValueError(
            f"Incorrect format for grid data in {input_file} with format {format}. "
            f"Expected 3 columns but found {len(data.colnames)}."
        )
    x0_col = data.colnames[0]
    x1_col = data.colnames[1]
    v_col = data.colnames[2]

    # Get the values along the x0 and x1 dimensions.
    x0 = np.sort(np.unique(data[x0_col].data))
    x1 = np.sort(np.unique(data[x1_col].data))

    # Get the array of values.
    if len(data) != len(x0) * len(x1):
        raise ValueError(
            f"Incomplete data for {input_file} with format {format}. Expected "
            f"{len(x0) * len(x1)} entries but found {len(data)}."
        )

    # If we are validating, loop through the entire table and check that
    # the x0 and x1 values are in the expected order.
    if validate:
        counter = 0
        for i in range(len(x0)):
            for j in range(len(x1)):
                if data[x0_col][counter] != x0[i]:
                    raise ValueError(
                        f"Incorrect x0 ordering in {input_file} at line={counter}."
                        f"Expected {x0[i]} but found {data[x0_col][counter]}."
                    )
                if data[x1_col][counter] != x1[j]:
                    raise ValueError(
                        f"Incorrect x1 ordering in {input_file} at line={counter}. "
                        f"Expected {x1[j]} but found {data[x1_col][counter]}."
                    )
                counter += 1

    # Build the values matrix.
    values = data[v_col].data.reshape((len(x0), len(x1)))

    return x0, x1, values


def _read_lclib_data_from_open_file(input_file):
    """Read SNANA's lclib data from a text file.

    Parameters
    ----------
    input_file : file
        The input data file containing SNANA's lclib data.

    Returns
    -------
    curves : list of astropy.table.Table
        A list of Astropy Tables, each representing a light curve.
    """
    colnames = []
    curves = []
    meta = {}
    current_model = {}
    parnames = []
    in_doc_block = False

    for l_num, line in enumerate(input_file):
        # Strip out the trailing comment. Then skip lines that are either
        # empty or do not contain a key-value pair.
        line = line.split("#")[0].strip()
        if not line or ":" not in line:
            continue

        # Split the line into key and value.
        key, value = line.split(":", 1)
        value = value.strip()

        # Handle the keys corresponding to a documentation block.
        if key == "DOCUMENTATION":
            in_doc_block = True
        elif key == "DOCUMENTATION_END":
            in_doc_block = False
        if in_doc_block:
            # If we are in a documentation block, just continue to the next line.
            continue

        if key == "COMMENT":
            continue  # Skip comments.
        elif key == "FILTERS":
            # Create a list of data columns with time and each filter.
            colnames = ["time"]
            for c in value:
                colnames.append(c)
        elif key == "END_EVENT":
            curr_id = meta.get("id", "")
            if curr_id != value:
                raise ValueError(f"Event ID mismatch (line {l_num}): found {value}, expected {curr_id}.")

            # Save the table we have so far.
            curves.append(Table(current_model, meta=meta))
        elif key == "START_EVENT":
            if len(colnames) == 0:
                raise ValueError(f"Error on line= {l_num}: No filters defined.")

            # Start a new light curve, but resetting the lists of data from the columns.
            current_model["type"] = []  # Initialize the type list.
            for col in colnames:
                current_model[col] = []
            meta["id"] = value
        elif key == "S" or key == "T":
            # Save an observation or template to the current light curve.
            current_model["type"].append(key)  # Get the type from the key.

            # Get the time and magnitudes from the columns.
            col_vals = value.split()
            if len(col_vals) != len(colnames):
                raise ValueError(f"Expected {len(colnames)} values on line={l_num}: {col_vals}")
            for col_idx, col in enumerate(colnames):
                current_model[col].append(float(col_vals[col_idx]))
        elif key == "MODEL_PARNAMES" or key == "MODEL_PARAMETERS":
            parnames = value.split(",")
        elif key == "PARVAL":
            if "," in value:
                all_vals = value.split(",")
            else:
                all_vals = value.split()

            if len(all_vals) != len(parnames):
                raise ValueError(f"Expected {len(parnames)} parameter values on line={l_num}: {all_vals}")
            meta["PARVAL"] = {key: value for key, value in zip(parnames, all_vals, strict=False)}
        else:
            # Save everything else to the meta dictionary.
            meta[key] = value

    return curves


def read_lclib_data(input_file):
    """Read SNANA's LCLIB data from a text file.

    Parameters
    ----------
    input_file : str or Path
        The path to the SNANA LCLIB data file.

    Returns
    -------
    curves : list of astropy.table.Table
        A list of Astropy Tables, each representing a light curve.
    """
    input_file = Path(input_file)
    logging.debug(f"Loading SNANA LCLIB data from {input_file}")
    if not input_file.is_file():
        raise FileNotFoundError(f"File {input_file} not found.")

    # Use the file suffix to determine how to read the file.
    suffix = input_file.suffix.lower()
    if suffix in [".gz", ".gzip"]:
        # Open as a gzipped text file.
        with gzip.open(input_file, "rt") as file_ptr:
            curves = _read_lclib_data_from_open_file(file_ptr)
    elif suffix in [".dat", ".txt", ".text"]:
        # Try to open the file as a regular text file.
        with open(input_file, "r") as file_ptr:
            curves = _read_lclib_data_from_open_file(file_ptr)
    else:  # pragma: no cover
        raise ValueError(f"Unsupported file format: {suffix}.")

    return curves


def read_sqlite_table(db_path, table_name=None, sql_query=None):
    """Read a table from a SQLite database into a pandas DataFrame.

    Parameters
    ----------
    db_path : str or Path
        The path to the SQLite database file.
    table_name : str, optional
        The name of the table to read. If not provided, sql_query must be provided.
    sql_query : str, optional
        A custom SQL query to execute. If provided, this query will be used instead of reading a table.

    Returns
    -------
    df : pandas.DataFrame
        The table data as a pandas DataFrame.
    """
    if table_name is None and sql_query is None:  # pragma: no cover
        raise ValueError("Either table_name or sql_query must be provided.")
    if sql_query is None:
        sql_query = f"SELECT * FROM {table_name}"

    logger = logging.getLogger(__name__)
    logger.debug(f"Reading table {table_name} from SQLite database {db_path}")

    db_path = Path(db_path)
    if not db_path.is_file():
        raise FileNotFoundError(f"Database file {db_path} not found.")

    # Connect to the SQLite database.
    con = sqlite3.connect(f"file:{db_path}?mode=ro", uri=True)

    # Read the specified table into a DataFrame.
    try:
        df = pd.read_sql_query(sql_query, con)
    except Exception as e:  # pragma: no cover
        raise RuntimeError(f"Error executing sql query '{sql_query}' from database {db_path}: {e}") from e

    # Close the database connection.
    con.close()

    return df



import numpy as np
import pytest
from lightcurvelynx.astro_utils.sed import SED
from lightcurvelynx.astro_utils.spectrograph import Spectrograph
from lightcurvelynx.models.static_sed_model import StaticBandfluxModel, StaticSEDModel
from lightcurvelynx.utils.extrapolate import LinearDecay
from lightcurvelynx.utils.io_utils import write_numpy_data


def test_single_static_sed() -> None:
    """Test that we can create and sample a StaticSEDModel object with a single SED."""
    sed = SED(
        wavelengths=np.array([100.0, 200.0, 300.0, 400.0]),
        fluxes=np.array([10.0, 20.0, 20.0, 10.0]),
    )
    model = StaticSEDModel(sed, node_label="test")
    assert len(model) == 1

    times = np.array([1, 2, 3, 10, 20])
    wavelengths = np.array([50.0, 100.0, 150.0, 200.0, 250.0, 300.0, 350.0, 400.0, 450.0])

    with pytest.warns(UserWarning):
        # Warns for wavelengths outside the SED range.
        values = model.evaluate_sed(times, wavelengths)
    assert values.shape == (5, 9)

    expected = np.array([0.0, 10.0, 15.0, 20.0, 20.0, 20.0, 15.0, 10.0, 0.0])
    for t_idx in range(5):
        assert np.array_equal(values[t_idx, :], expected)


def test_single_static_sed_from_numpy() -> None:
    """Test that we can create and sample a StaticSEDModel object with a single SED."""
    sed = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    model = StaticSEDModel(sed, node_label="test")
    assert len(model) == 1

    times = np.array([1, 2, 3, 10, 20])
    wavelengths = np.array([50.0, 100.0, 150.0, 200.0, 250.0, 300.0, 350.0, 400.0, 450.0])

    with pytest.warns(UserWarning):
        # Warns for wavelengths outside the SED range.
        values = model.evaluate_sed(times, wavelengths)
    assert values.shape == (5, 9)

    expected = np.array([0.0, 10.0, 15.0, 20.0, 20.0, 20.0, 15.0, 10.0, 0.0])
    for t_idx in range(5):
        assert np.array_equal(values[t_idx, :], expected)

    # We can evaluate spectral (even if they are sampled differently).
    sg_pbg = Spectrograph.from_regular_grid(wave_start=150.0, wave_end=475.0, bin_width=50.0)
    with pytest.warns(UserWarning):
        # Warns for wavelengths outside the SED range.
        values2 = model.evaluate_spectra(times, sg_pbg, None)
    expected_row = np.array([17.5, 20.0, 20.0, 17.5, 12.5, 0.0])
    expected_all = np.tile(expected_row, (5, 1))
    assert np.allclose(values2, expected_all)


def test_static_sed_fail() -> None:
    """Test that we correctly fail on bad SEDs."""
    # Non-numpy arrays
    with pytest.raises(ValueError):
        _ = StaticSEDModel([None], node_label="test")
    with pytest.raises(ValueError):
        _ = StaticSEDModel([1.0], node_label="test")

    # Not enough points.
    sed = np.array([[100.0], [10.0], [0.0]])
    with pytest.raises(ValueError):
        _ = StaticSEDModel([sed], node_label="test")

    # Incorrectly shaped data.
    sed = np.array(
        [
            [100.0, 200.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0],  # fluxes
            [0.0, 0.0, 0.0],  # Other row
        ]
    )
    with pytest.raises(ValueError):
        _ = StaticSEDModel([sed], node_label="test")

    # Wavelengths unsorted.
    sed = np.array(
        [
            [100.0, 200.0, 400.0, 300.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    with pytest.raises(ValueError):
        _ = StaticSEDModel([sed], node_label="test")


def test_static_sed_from_file(tmp_path) -> None:
    """Test that we can create a StaticSEDModel object from a file."""
    test_sed = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    times = np.array([1, 2, 3, 10, 20])
    wavelengths = np.array([100.0, 150.0, 200.0, 250.0, 300.0, 350.0, 400.0])
    expected = np.array([10.0, 15.0, 20.0, 20.0, 20.0, 15.0, 10.0])

    for fmt in ["npy", "npz", "txt", "csv"]:
        file_path = tmp_path / f"test_sed.{fmt}"
        write_numpy_data(file_path, test_sed.T)

        model = StaticSEDModel.from_file(file_path, node_label="test")
        assert len(model) == 1

        values = model.evaluate_sed(times, wavelengths)
        assert values.shape == (5, 7)

        for t_idx in range(5):
            assert np.array_equal(values[t_idx, :], expected)

    # Try an invalid array shape.
    test_sed_invalid = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
            [0.0, 0.0, 0.0, 0.0],  # Other row
        ]
    )
    file_path_invalid = tmp_path / "test_sed.csv"
    write_numpy_data(file_path_invalid, test_sed_invalid.T)

    with pytest.raises(ValueError):
        _ = StaticSEDModel.from_file(file_path_invalid, node_label="test")


def test_multiple_static_seds() -> None:
    """Test that we can create and sample a StaticSEDModel object with a multiple SEDs."""
    # We use a combination of numpy arrays and SED objects.
    sed0 = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    sed1 = SED(
        wavelengths=np.array([100.0, 200.0, 300.0, 400.0]),
        fluxes=np.array([20.0, 40.0, 40.0, 20.0]),
    )
    model = StaticSEDModel([sed0, sed1], weights=[0.25, 0.75], node_label="test")
    assert len(model) == 2

    # Check that all of the indices are 0 or 1 and the split is approximately 25/75
    params = model.sample_parameters(num_samples=10_000)
    inds_0 = params["test"]["selected_idx"] == 0
    inds_1 = params["test"]["selected_idx"] == 1
    assert np.all(inds_0 | inds_1)
    assert 1_500 < np.count_nonzero(inds_0) < 3_500
    assert 6_500 < np.count_nonzero(inds_1) < 8_500

    times = np.array([1, 2, 3])
    wavelengths = np.array([50.0, 100.0, 150.0, 200.0, 250.0, 300.0, 350.0, 400.0, 450.0])
    with pytest.warns(UserWarning):
        # Warns for wavelengths outside the SED range.
        values = model.evaluate_sed(times, wavelengths, params)

    expected_0 = np.tile([0.0, 10.0, 15.0, 20.0, 20.0, 20.0, 15.0, 10.0, 0.0], 3).reshape(3, 9)
    expected_1 = 2.0 * expected_0

    for idx in range(10_000):
        if inds_0[idx]:
            assert np.array_equal(values[idx], expected_0)
        else:
            assert np.array_equal(values[idx], expected_1)

    # We can retrieve SEDs by index and iterate over them. Everything is
    # extracted as an SED object.
    assert isinstance(model[0], SED)
    assert isinstance(model[1], SED)

    count = 0
    for sed in model:
        assert isinstance(sed, SED)
        count += 1
    assert count == 2


def test_multiple_static_seds_min_max():
    """Test that we can get the min and max wavelengths from a StaticSEDModel object with multiple SEDs."""
    sed0 = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    sed1 = np.array(
        [
            [200.0, 300.0, 400.0, 500.0],  # Wavelengths
            [20.0, 40.0, 40.0, 20.0],  # fluxes
        ]
    )
    model = StaticSEDModel([sed0, sed1], weights=[0.5, 0.5], node_label="test")

    states = model.sample_parameters(num_samples=1)

    # Force the selected_idx to be 0 for the test.
    states.set("test", "selected_idx", 0)
    assert model.minwave(states) == 100.0
    assert model.maxwave(states) == 400.0

    # Force the selected_idx to be 1 for the test.
    states.set("test", "selected_idx", 1)
    assert model.minwave(states) == 200.0
    assert model.maxwave(states) == 500.0

    # We fail if we do not pass in the states.
    with pytest.raises(ValueError):
        _ = model.minwave()
    with pytest.raises(ValueError):
        _ = model.maxwave()


def test_single_static_bandflux() -> None:
    """Test that we can create and sample a StaticBandfluxModel object with a single bandflux."""
    bandflux = {"r": 10.0, "g": 20.0, "b": 30.0}
    model = StaticBandfluxModel(bandflux, node_label="test")
    assert len(model) == 1
    assert len(model.list_effects()) == 0

    times = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    filters = ["r", "r", "g", "b", "r", "g", "b", "g", "b", "r"]
    expected = np.array([10.0, 10.0, 20.0, 30.0, 10.0, 20.0, 30.0, 20.0, 30.0, 10.0])

    state = model.sample_parameters(num_samples=1)
    fluxes = model.evaluate_bandfluxes(None, times, filters, state)
    assert len(fluxes) == 10
    assert np.array_equal(fluxes, expected)

    # We fail if we try to evaluate a spectra from a bandflux only model.
    sg_pbg = Spectrograph.from_regular_grid(wave_start=150.0, wave_end=475.0, bin_width=50.0)
    with pytest.raises(NotImplementedError):
        _ = model.evaluate_spectra(times, sg_pbg, state)


def test_multiple_static_bandflux() -> None:
    """Test that we can create and sample a StaticBandfluxModel object with multiple bandfluxes."""
    bandflux0 = {"r": 10.0, "g": 20.0, "b": 30.0}
    bandflux1 = {"r": 15.0, "g": 25.0, "b": 35.0}
    model = StaticBandfluxModel([bandflux0, bandflux1], weights=[0.25, 0.75], node_label="test")
    assert len(model) == 2

    # Check that all of the indices are 0 or 1 and the split is approximately 25/75
    params = model.sample_parameters(num_samples=10_000)
    inds_0 = params["test"]["selected_idx"] == 0
    inds_1 = params["test"]["selected_idx"] == 1
    assert np.all(inds_0 | inds_1)
    assert 1_500 < np.count_nonzero(inds_0) < 3_500
    assert 6_500 < np.count_nonzero(inds_1) < 8_500

    times = np.array([1, 2, 3, 4, 5])
    filters = ["r", "r", "g", "b", "r"]
    expected0 = np.array([10.0, 10.0, 20.0, 30.0, 10.0])
    expected1 = np.array([15.0, 15.0, 25.0, 35.0, 15.0])

    fluxes = model.evaluate_bandfluxes(None, times, filters, params)
    assert fluxes.shape == (10_000, 5)
    for idx in range(10_000):
        if inds_0[idx]:
            assert np.array_equal(fluxes[idx], expected0)
        else:
            assert np.array_equal(fluxes[idx], expected1)


def test_static_bandflux_set_extrapolation():
    """Test that we can set the time extrapolation for a StaticBandfluxModel."""
    bandflux = {"r": 10.0, "g": 20.0, "b": 30.0}

    before_extrap = LinearDecay(decay_width=10.0)  # 10 days to go to zero
    after_extrap = LinearDecay(decay_width=20.0)  # 20 days to go to zero
    model = StaticBandfluxModel(bandflux, time_extrapolation=(before_extrap, after_extrap), node_label="test")
    assert model._time_extrap_before is not None
    assert model._time_extrap_after is not None
    assert isinstance(model._time_extrap_before, LinearDecay)
    assert isinstance(model._time_extrap_after, LinearDecay)

    # We provide a warning if the user tries to set wave_extrapolation.
    with pytest.warns(UserWarning):
        _ = StaticBandfluxModel(bandflux, wave_extrapolation=LinearDecay(decay_width=5.0))

    # We fail with invalid time_extrapolation settings.
    with pytest.raises(ValueError):
        _ = StaticBandfluxModel(bandflux, time_extrapolation=(before_extrap, before_extrap, before_extrap))
    with pytest.raises(TypeError):
        _ = StaticBandfluxModel(bandflux, time_extrapolation="not a model")


class DummySynphotModel:
    """A fake synphot model used for testing.

    Attributes
    ----------
    waveset : numpy.ndarray
        The wavelengths at which the SED is defined (in angstroms)
    fluxset : numpy.ndarray
        The flux at each given wavelength (in nJy.)
    """

    def __init__(self, waveset, fluxset):
        self.waveset = waveset
        self.fluxset = fluxset
        self.z = 0.0  # Redshift

    def __call__(self, waves, **kwargs):
        """Return the flux for the given wavelengths as interpolated PHOTLAM.

        Parameters
        ----------
        waves : numpy.ndarray
            The wavelengths at which to evaluate the SED (in angstroms).
        **kwargs : dict
            Additional keyword arguments (ignored).

        Returns
        -------
        numpy.ndarray
            The interpolated flux values at the given wavelengths (in PHOTLAM).
        """
        # Return a dummy SED for the given wavelengths
        return np.interp(waves, self.waveset, self.fluxset, left=0.0, right=0.0)


def test_static_sed_from_synphot() -> None:
    """Test that we can create a StaticSEDModel from a synphot model."""
    # Create a dummy model with 4 samples of SEDs [10.0, 20.0, 30.0, 40.0] in nJy.
    # Since synphot uses PHOTLAM, we preconvert and provide in that unit.
    sp_model = DummySynphotModel(
        waveset=np.array([1000.0, 2000.0, 3000.0, 4000.0]),
        fluxset=np.array([1.50919018e-08, 1.50919018e-08, 1.50919018e-08, 1.50919018e-08]),
    )
    model = StaticSEDModel.from_synphot(sp_model)
    assert len(model) == 1

    times = np.array([1, 2, 3, 10, 20])
    wavelengths = np.array([500.0, 1000.0, 1500.0, 2000.0, 3000.0, 5000.0])
    expected = np.array([0.0, 10.0, 15.0, 20.0, 30.0, 0.0])
    with pytest.warns(UserWarning):
        # Warns for wavelengths outside the SED range.
        fluxes = model.evaluate_sed(times, wavelengths)
    assert fluxes.shape == (len(times), len(wavelengths))
    for i in range(len(times)):
        np.testing.assert_allclose(fluxes[i, :], expected, rtol=1e-5)

    # We fail is the synphot model has a redshift defined.
    sp_model.z = 0.5
    with pytest.raises(ValueError):
        _ = StaticSEDModel.from_synphot(sp_model)


def test_static_bandflux_with_extrapolation():
    """Test that we can extrapolate a Bandflux model by adding time bounds
    to a StaticBandfluxModel.
    """

    class _BoundStaticBandfluxModel(StaticBandfluxModel):
        """A StaticBandfluxModel that defines time bounds for extrapolation."""

        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self._time_bounds = (0.0, 10.0)

        def minphase(self, **kwargs):
            """Return the minimum phase for the model."""
            return self._time_bounds[0]

        def maxphase(self, **kwargs):
            """Return the maximum phase for the model."""
            return self._time_bounds[1]

    before_extrap = LinearDecay(decay_width=10.0)  # 10 days to go to zero
    after_extrap = LinearDecay(decay_width=20.0)  # 20 days to go to zero
    bandflux = {"r": 10.0, "g": 20.0, "b": 30.0}
    model = _BoundStaticBandfluxModel(
        bandflux,
        time_extrapolation=(before_extrap, after_extrap),
        node_label="test",
    )

    # Data in the form of (time, filter, expected_flux)
    data = np.array(
        [
            [-5.0, "r", 5.0],  # 50% of red
            [-4.0, "r", 6.0],  # 60% of red
            [-2.0, "g", 16.0],  # 80% of green
            [-1.0, "b", 27.0],  # 90% of blue
            [0.0, "r", 10.0],  # No extrapolation
            [1.0, "g", 20.0],  # No extrapolation
            [2.0, "b", 30.0],  # No extrapolation
            [3.0, "r", 10.0],  # No extrapolation
            [4.0, "g", 20.0],  # No extrapolation
            [5.0, "b", 30.0],  # No extrapolation
            [5.0, "b", 30.0],  # No extrapolation
            [11.0, "r", 9.5],  # 95% of red
            [12.0, "g", 18.0],  # 90% of green
            [13.0, "b", 25.5],  # 85% of blue
            [14.0, "r", 8.0],  # 80% of red
            [15.0, "b", 22.5],  # 75% of blue
        ]
    )

    times = data[:, 0].astype(float)
    filters = data[:, 1]
    expected = data[:, 2].astype(float)
    state = model.sample_parameters(num_samples=1)
    fluxes = model.evaluate_bandfluxes(None, times, filters, state)
    assert np.array_equal(fluxes, expected)



import numpy as np
import pytest
from lightcurvelynx.astro_utils.sed import SED
from lightcurvelynx.utils.io_utils import write_numpy_data


def test_create_sed() -> None:
    """Test that we can create and sample a StaticSEDModel object with a single SED."""
    sed = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    model = SED(sed[0], sed[1])
    assert model.minwave() == 100.0
    assert model.maxwave() == 400.0

    # We can interpolate the SED at given wavelengths.
    wavelengths = np.array([50.0, 100.0, 150.0, 200.0, 250.0, 300.0, 350.0, 400.0, 450.0])
    values = model.evaluate(wavelengths)
    assert values.shape == (9,)

    expected = np.array([0.0, 10.0, 15.0, 20.0, 20.0, 20.0, 15.0, 10.0, 0.0])
    np.testing.assert_array_equal(values, expected)


def test_sed_fail() -> None:
    """Test that we correctly fail on bad SEDs."""
    with pytest.raises(ValueError):
        _ = SED([1.0, 2.0], [1.0])  # Different lengths
    with pytest.raises(ValueError):
        _ = SED([1.0], [1.0])  # Too few points
    with pytest.raises(ValueError):
        _ = SED([2.0, 1.0], [1.0, 2.0])  # Unsorted wavelengths


def test_sed_to_from_file(tmp_path) -> None:
    """Test that we can create a StaticSEDModel object from a file."""
    test_sed = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    wavelengths = np.array([100.0, 150.0, 200.0, 250.0, 300.0, 350.0, 400.0])
    expected = np.array([10.0, 15.0, 20.0, 20.0, 20.0, 15.0, 10.0])
    sed_org = SED(test_sed[0], test_sed[1])

    for fmt in ["npy", "npz", "txt", "csv"]:
        file_path = tmp_path / f"test_sed_file.{fmt}"
        assert not file_path.exists()

        sed_org.to_file(file_path)
        assert file_path.exists()

        model = SED.from_file(file_path)
        values = model.evaluate(wavelengths)
        assert np.allclose(values, expected)

    # Try reading an invalid array shape.
    test_sed_invalid = np.array(
        [
            [100.0, 200.0, 300.0, 400.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
            [0.0, 0.0, 0.0, 0.0],  # Other row
        ]
    )
    file_path_invalid = tmp_path / "test_invalid_sed.csv"
    write_numpy_data(file_path_invalid, test_sed_invalid.T)

    with pytest.raises(ValueError):
        _ = SED.from_file(file_path_invalid)


class DummySynphotModel:
    """A fake synphot model used for testing.

    Attributes
    ----------
    waveset : numpy.ndarray
        The wavelengths at which the SED is defined (in angstroms)
    fluxset : numpy.ndarray
        The flux at each given wavelength (in nJy.)
    """

    def __init__(self, waveset, fluxset):
        self.waveset = waveset
        self.fluxset = fluxset
        self.z = 0.0  # Redshift

    def __call__(self, waves, **kwargs):
        """Return the flux for the given wavelengths as interpolated PHOTLAM.

        Parameters
        ----------
        waves : numpy.ndarray
            The wavelengths at which to evaluate the SED (in angstroms).
        **kwargs : dict
            Additional keyword arguments (ignored).

        Returns
        -------
        numpy.ndarray
            The interpolated flux values at the given wavelengths (in PHOTLAM).
        """
        # Return a dummy SED for the given wavelengths
        return np.interp(waves, self.waveset, self.fluxset, left=0.0, right=0.0)


def test_sed_from_synphot() -> None:
    """Test that we can create a SED from a synphot model."""
    # Create a dummy model with 4 samples of SEDs [10.0, 20.0, 30.0, 40.0] in nJy.
    # Since synphot uses PHOTLAM, we preconvert and provide in that unit.
    sp_model = DummySynphotModel(
        waveset=np.array([1000.0, 2000.0, 3000.0, 4000.0]),
        fluxset=np.array([1.50919018e-08, 1.50919018e-08, 1.50919018e-08, 1.50919018e-08]),
    )
    model = SED.from_synphot(sp_model)

    wavelengths = np.array([500.0, 1000.0, 1500.0, 2000.0, 3000.0, 5000.0])
    expected = np.array([0.0, 10.0, 15.0, 20.0, 30.0, 0.0])
    fluxes = model.evaluate(wavelengths)
    assert fluxes.shape == (len(wavelengths),)
    np.testing.assert_allclose(fluxes, expected, rtol=1e-5)


@pytest.mark.filterwarnings("ignore::UserWarning")
def test_sed_plot() -> None:
    """Test that we can plot a SED."""
    sed = np.array(
        [
            [1000.0, 2000.0, 3000.0, 4000.0],  # Wavelengths
            [10.0, 20.0, 20.0, 10.0],  # fluxes
        ]
    )
    model = SED(sed[0], sed[1])
    model.plot()  # Just ensure this runs without error



import numpy as np
import pytest
from lightcurvelynx.astro_utils.spectrograph import Spectrograph


def test_create_spectrograph_from_regular_grid():
    """Test that we can create and query a Spectrograph object."""
    spgraph = Spectrograph.from_regular_grid(wave_start=3000, wave_end=11000, bin_width=5.0)
    assert spgraph.instrument == "Spectrograph"
    assert len(spgraph) == 1600  # (11000 - 3000) / 5 = 1600 bins
    assert np.array_equal(spgraph.waves, np.arange(3000 + 2.5, 11000, 5.0))

    l_val, h_val = spgraph.wave_bounds()
    assert l_val == 3000
    assert h_val == 11000

    for i in range(len(spgraph)):
        assert spgraph.bin_width(i) == pytest.approx(5.0)

    assert str(spgraph) == "Spectrograph (spectra) [3000.0A - 11000.0A]"

    # Two dimensional fluxes to spec_fluxes
    values = np.random.random((10, len(spgraph)))
    spec_fluxes = spgraph.evaluate(values)
    assert np.allclose(spec_fluxes, values)

    # Three dimensional fluxes to spec_fluxes
    values_3d = np.random.random((4, 10, len(spgraph)))
    spec_fluxes_3d = spgraph.evaluate(values_3d)
    assert np.allclose(spec_fluxes_3d, values_3d)

    # Test that we fail to create a Spectrograph object with invalid parameters.
    with pytest.raises(ValueError):
        _ = Spectrograph.from_regular_grid(wave_start=5000, wave_end=4000, bin_width=5.0)
    with pytest.raises(ValueError):
        _ = Spectrograph.from_regular_grid(wave_start=4000, wave_end=5000, bin_width=-5.0)


def test_create_spectrograph_from_irregular_grid():
    """Test that we can create and query a Spectrograph object."""
    waves = np.array([3500.0, 4000.0, 5000.0, 7000.0, 7500.0, 8000.0])
    spgraph = Spectrograph(waves, instrument="custom_spectrograph")
    assert spgraph.instrument == "custom_spectrograph"
    assert len(spgraph) == 6
    assert np.array_equal(spgraph.waves, waves)

    l_val, h_val = spgraph.wave_bounds()
    assert l_val == 3250.0
    assert h_val == 8250.0

    assert spgraph.bin_width(0) == pytest.approx(500.0)
    assert spgraph.bin_width(1) == pytest.approx(750.0)
    assert spgraph.bin_width(2) == pytest.approx(1500.0)
    assert spgraph.bin_width(3) == pytest.approx(1250.0)
    assert spgraph.bin_width(4) == pytest.approx(500.0)
    assert spgraph.bin_width(5) == pytest.approx(500.0)

    assert str(spgraph) == "custom_spectrograph (spectra) [3250.0A - 8250.0A]"
    # Two dimensional fluxes to spec_fluxes
    values = np.random.random((10, len(spgraph)))
    spec_fluxes = spgraph.evaluate(values)
    assert np.allclose(spec_fluxes, values)

    # Three dimensional fluxes to spec_fluxes
    values_3d = np.random.random((4, 10, len(spgraph)))
    spec_fluxes_3d = spgraph.evaluate(values_3d)
    assert np.allclose(spec_fluxes_3d, values_3d)

    # Test that we fail to create a Spectrograph object with invalid parameters.
    bad_waves = np.array([4000.0, 3500.0, 5000.0])
    with pytest.raises(ValueError):
        _ = Spectrograph(bad_waves)


def test_spectrograph_equals():
    """Test that we can compare two Spectrograph objects for equality."""
    spgraph1 = Spectrograph.from_regular_grid(wave_start=4000, wave_end=8000, bin_width=5.0)
    spgraph2 = Spectrograph.from_regular_grid(wave_start=4000, wave_end=8000, bin_width=5.0)
    spgraph3 = Spectrograph.from_regular_grid(wave_start=4000, wave_end=8000, bin_width=10.0)
    spgraph4 = Spectrograph.from_regular_grid(wave_start=3000, wave_end=8000, bin_width=5.0)
    spgraph5 = Spectrograph.from_regular_grid(wave_start=4000, wave_end=9000, bin_width=5.0)

    assert spgraph1 == spgraph2
    assert spgraph1 != spgraph3
    assert spgraph1 != spgraph4
    assert spgraph1 != spgraph5


def test_create_spectrograph_with_scale():
    """Test that we can create and query a Spectrograph object."""
    scale = np.array([0.5, 1.0, 1.0, 1.0, 0.8])
    spgraph = Spectrograph.from_regular_grid(wave_start=4000, wave_end=5000, bin_width=200.0, scale=scale)
    assert np.allclose(spgraph.waves, np.array([4100.0, 4300.0, 4500.0, 4700.0, 4900.0]))

    # Two dimensional fluxes to spec_fluxes
    input = np.array(
        [
            [10.0, 20.0, 30.0, 40.0, 50.0],
            [5.0, 15.0, 25.0, 35.0, 45.0],
        ]
    )
    expected = np.array(
        [
            [5.0, 20.0, 30.0, 40.0, 40.0],
            [2.5, 15.0, 25.0, 35.0, 36.0],
        ]
    )
    results = spgraph.evaluate(input)
    assert np.allclose(results, expected)

    # Three dimensional fluxes to bandfluxes
    input = np.array(
        [
            [
                [10.0, 20.0, 30.0, 40.0, 50.0],
                [5.0, 15.0, 25.0, 35.0, 45.0],
            ],
            [
                [8.0, 18.0, 28.0, 38.0, 48.0],
                [4.0, 14.0, 24.0, 34.0, 44.0],
            ],
        ]
    )
    expected = np.array(
        [
            [
                [5.0, 20.0, 30.0, 40.0, 40.0],
                [2.5, 15.0, 25.0, 35.0, 36.0],
            ],
            [
                [4.0, 18.0, 28.0, 38.0, 38.4],
                [2.0, 14.0, 24.0, 34.0, 35.2],
            ],
        ]
    )
    results = spgraph.evaluate(input)
    assert np.allclose(results, expected)

    # Test equality with scales.
    waves = np.copy(spgraph.waves)
    spgraph2 = Spectrograph(waves, scale=scale)
    assert spgraph == spgraph2

    different_scale = np.array([0.5, 1.0, 1.0, 1.0, 0.9])
    spgraph3 = Spectrograph(waves, scale=different_scale)
    assert spgraph != spgraph3

    # Test with a mismatched scale length.
    bad_scale = np.array([1.0, 0.8])
    with pytest.raises(ValueError):
        _ = Spectrograph(waves, scale=bad_scale)



"""Wrappers for the models defined in bagle.

https://github.com/MovingUniverseLab/BAGLE_Microlensing
"""

from citation_compass import CiteClass

from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.math_nodes.given_sampler import TableSampler
from lightcurvelynx.models.physical_model import BandfluxModel


def _load_bagle_model_class(model_name):
    """Load a bagle model class by name.

    Parameters
    ----------
    model_name : str
        The name of the bagle model class to load.

    Returns
    -------
    model_class : class
        The bagle model class.
    """
    try:
        from bagle import model
    except ImportError as err:  # pragma: no cover
        raise ImportError(
            "The bagle package is required to use the BagleWrapperModel. Please install it. "
            "See https://bagle.readthedocs.io/en/latest/installation.html for instructions."
        ) from err
    model_class = getattr(model, model_name)
    return model_class


class BagleWrapperModel(BandfluxModel, CiteClass):
    """A wrapper for single bagle models (one model type).

    Parameterized values include:
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The t0 of the zero phase, date. [from BasePhysicalModel]
    Additional parameterized values are used for specific bagle models.

    Note
    ----
    The t0 parameter saved in the results may be approximate depending on the bagle model
    used. Some models compute t0 from other parameters (e.g., time of closest approach).
    This updated t0 is not saved in the results.

    References
    ----------
    * Lu et al., “The BAGLE Python Package for Bayesian Analysis of Gravitational Lensing Events”,
      AAS Journals, submitted
    * Bhadra et al., “Modeling Binary Lenses and Sources with the BAGLE Python Package”, AAS Journals,
      submitted
    * Chen et al., “Adjusting Gaussian Process Priors for BAGLE's Gravitational Microlensing Model Fits”,
      in prep.

    Parameters
    ----------
    model_info : str or class
        The name of the bagle model class to use in the simulation or the class itself.
    parameter_dict : dict
        A dictionary of parameter names and values to use for the model. The keys should
        match the parameter names expected by the bagle model.
    filter_idx : dict, optional
        A mapping from filter names to indices expected by the bagle model. If not provided,
        a default mapping for ugrizy filters to [0, 1, 2, 3, 4, 5] will be used.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    # Convenience mapping from filter name to index in the parameter list.
    _default_filter_idx = {"u": 0, "g": 1, "r": 2, "i": 3, "z": 4, "y": 5}

    def __init__(self, model_info, parameter_dict, filter_idx=None, **kwargs):
        # We start by extracting the parameter information needed for a general physical model.
        # We check the parameter dictionary first, falling back to kwargs if needed.
        ra = parameter_dict.get("raL", None)
        if "ra" in kwargs:
            if ra is None:
                ra = kwargs.pop("ra")
            else:
                raise ValueError(
                    "The 'ra' parameter is specified in both parameter_dict (as 'raL') "
                    " and kwargs (as 'ra'). Please only use the parameter_dict."
                )

        dec = parameter_dict.get("decL", None)
        if "dec" in kwargs:
            if dec is None:
                dec = kwargs.pop("dec")
            else:
                raise ValueError(
                    "The 'dec' parameter is specified in both parameter_dict (as 'decL') "
                    " and kwargs (as 'dec'). Please only use the parameter_dict."
                )

        # The t0 parameter can be specified different ways from the bagle model,
        # including being computed from other parameters. We need a base value to use
        # for the time bounds, so we use t0 if it is in the parameter_dict or otherwise
        # anchor on any parameter that starts with "t0" (under the assumption that it will
        # indicate the same general range of times).
        t0 = parameter_dict.get("t0", None)
        if t0 is None:
            for key, value in parameter_dict.items():
                if key.startswith("t0"):
                    t0 = value
                    break
        if "t0" in kwargs:
            raise ValueError(
                "The 't0' parameter must be specified in only the 'parameter_dict' for the BagleWrapperModel."
            )

        super().__init__(ra=ra, dec=dec, t0=t0, **kwargs)

        # Add all of the parameters in the dictionary as settable parameters (if they are not
        # already set by the parent class) and save their names (in order) for later use.
        self._parameter_names = []
        for param_name, param_value in parameter_dict.items():
            self._parameter_names.append(param_name)
            if param_name in self.list_params():
                self.set_parameter(param_name, param_value)
            else:
                self.add_parameter(param_name, param_value)

        # Save the model class, but DO NOT create the model object yet. We allow the
        # user to pass in a class to simplify testing when bagle is not installed.
        if isinstance(model_info, str):
            self._model_class = _load_bagle_model_class(model_info)
        else:
            self._model_class = model_info

        # Save the filter index mapping, using the default if none is provided.
        if filter_idx is None:
            self._filter_idx = self._default_filter_idx
        else:
            self._filter_idx = filter_idx

    @property
    def parameter_names(self):
        """The names of the parameters for this model."""
        return self._parameter_names

    def compute_bandflux(self, times, filter, state):
        """Evaluate the model at the passband level for a single, given graph state and filter.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filter : str
            The name of the filter.
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
            This is not used in this model, but is required for the function signature.

        Returns
        -------
        bandflux : numpy.ndarray
            A length T array of band fluxes for this model in this filter.
        """
        # Extract the local parameters for this object from the full state object.
        local_params = self.get_local_params(state)

        # Create the bagle model object and set the parameters from the current state. We do this
        # here because the parameters saved in `state` will be different in each run.
        current_params = {param_name: local_params[param_name] for param_name in self.parameter_names}
        model_obj = self._model_class(**current_params)

        # Use the newly created model object with the current parameters to compute the photometry.
        mags = model_obj.get_photometry(times, self._filter_idx[filter])
        bandflux = mag2flux(mags)
        return bandflux


class BagleMultiWrapperModel(BandfluxModel, CiteClass):
    """A wrapper for multiple bagle models (multiple model types).

    Parameterized values include:
      * dec - The object's declination in degrees. [from BasePhysicalModel]
      * distance - The object's luminosity distance in pc. [from BasePhysicalModel]
      * ra - The object's right ascension in degrees. [from BasePhysicalModel]
      * redshift - The object's redshift. [from BasePhysicalModel]
      * t0 - The t0 of the zero phase, date. [from BasePhysicalModel]
    Additional parameterized values are used for specific bagle models.

    References
    ----------
    * Lu et al., “The BAGLE Python Package for Bayesian Analysis of Gravitational Lensing Events”,
      AAS Journals, submitted
    * Bhadra et al., “Modeling Binary Lenses and Sources with the BAGLE Python Package”, AAS Journals,
      submitted
    * Chen et al., “Adjusting Gaussian Process Priors for BAGLE's Gravitational Microlensing Model Fits”,
      in prep.

    Atttributes
    -----------
    num_models : int
        The number of models being wrapped.
    parameter_dicts : dict
        A list of parameter dictionaries, one per model, each containing the parameter names
        and values for use in the corresponding model.

    Parameters
    ----------
    models : list of str or class
        The bagle model classes (or their names as strings) to use in the simulation.
    parameter_dicts : dict
        A list of parameter dictionaries, one per model, each containing.
    filter_idx : dict, optional
        A mapping from filter names to indices expected by the bagle model. If not provided,
        a default mapping for ugrizy filters to [0, 1, 2, 3, 4, 5] will be used.
    in_order : bool
        Return the given data in order of the rows (True). If False, performs
        random sampling with replacement. Default: False
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    # Convenience mapping from filter name to index in the parameter list.
    _default_filter_idx = {"u": 0, "g": 1, "r": 2, "i": 3, "z": 4, "y": 5}

    def __init__(self, models, parameter_dicts, filter_idx=None, in_order=False, **kwargs):
        self.parameter_dicts = parameter_dicts.copy()
        if len(models) != len(parameter_dicts):
            raise ValueError("The number of models must match the number of parameter dictionaries.")
        self.num_models = len(models)

        # We start by building a list of all the parameters across all the models.
        all_params = set()
        for param_dict in parameter_dicts:
            all_params.update(param_dict.keys())
        all_params = list(all_params)
        all_params.sort()

        # Next, we build a table of the model information (names and parameters). We also handle the
        # different t0 labels here, extracting the first one we find for use in the base physical model.
        sampler_data = {
            "_model_info": models,
            "_model_idx": list(range(len(models))),
            "_model_t0": [None] * len(models),
        }
        for param in all_params:
            sampler_data[param] = []
            for idx in range(self.num_models):
                value = parameter_dicts[idx].get(param, None)
                sampler_data[param].append(value)

                # Check for t0 or t0-like parameters to use for the base physical model.
                if param.startswith("t0") and sampler_data["_model_t0"][idx] is None:
                    sampler_data["_model_t0"][idx] = value
        self.sampler = TableSampler(sampler_data, in_order=in_order)

        # For the base physical model, we need to extract ra, dec, and t0. We check the parameter
        # dictionaries first, falling back to kwargs if needed.
        if "raL" in sampler_data:
            ra = self.sampler.raL
            if "ra" in kwargs:
                raise ValueError(
                    "The 'ra' parameter is specified in both parameter_dicts (as 'raL') "
                    " and kwargs (as 'ra'). Please only use the parameter_dicts."
                )
        elif "ra" in kwargs:
            ra = kwargs.pop("ra")
        else:
            ra = None

        if "decL" in sampler_data:
            dec = self.sampler.decL
            if "dec" in kwargs:
                raise ValueError(
                    "The 'dec' parameter is specified in both parameter_dicts (as 'decL') "
                    " and kwargs (as 'dec'). Please only use the parameter_dicts."
                )
        elif "dec" in kwargs:
            dec = kwargs.pop("dec")
        else:
            dec = None

        if "t0" in kwargs:
            raise ValueError(
                "The 't0' parameter must be specified in only the 'parameter_dict' "
                " for the BagleMultiWrapperModel."
            )

        super().__init__(ra=ra, dec=dec, t0=self.sampler._model_t0, **kwargs)

        # Add all of the parameters in the dictionary as settable parameters (if they are not
        # already set by the parent class).
        for param_name in sampler_data:
            if param_name in self.list_params():
                self.set_parameter(param_name, self.sampler.get_parameter_indicator(param_name))
            else:
                self.add_parameter(param_name, self.sampler.get_parameter_indicator(param_name))

        # Save the filter index mapping, using the default if none is provided.
        if filter_idx is None:
            self._filter_idx = self._default_filter_idx
        else:
            self._filter_idx = filter_idx

    def __len__(self):
        """The number of models wrapped by this object."""
        return self.num_models

    def compute_bandflux(self, times, filter, state):
        """Evaluate the model at the passband level for a single, given graph state and filter.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filter : str
            The name of the filter.
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
            This is not used in this model, but is required for the function signature.

        Returns
        -------
        bandflux : numpy.ndarray
            A length T array of band fluxes for this model in this filter.
        """
        # Extract the local parameters for this object from the full state object.
        local_params = self.get_local_params(state)

        # Get the class of the model to use.
        model_info = local_params["_model_info"]
        if isinstance(model_info, str):
            model_class = _load_bagle_model_class(model_info)
        else:
            model_class = model_info

        # Get the model index and the list of keys.
        model_idx = local_params["_model_idx"]
        model_keys = self.parameter_dicts[model_idx].keys()

        # Create the bagle model object and set the parameters from the current state. We do this
        # here because the parameters saved in `state` will be different in each run.
        current_params = {}
        for param_name in model_keys:
            value = local_params.get(param_name, None)
            if value is not None:
                current_params[param_name] = value
        model_obj = model_class(**current_params)

        # Use the newly created model object with the current parameters to compute the photometry.
        mags = model_obj.get_photometry(times, self._filter_idx[filter])
        bandflux = mag2flux(mags)
        return bandflux



"""A general extinction effect that wraps multiple backend libraries including:
* dust_extinction (https://github.com/karllark/dust_extinction)
* sncosmo extinction (https://github.com/sncosmo/extinction)
"""

import importlib
from pkgutil import iter_modules

import astropy.units as u
import numpy as np
from citation_compass import CiteClass

from lightcurvelynx.effects.effect_model import EffectModel


class _DustExtinctionWrapper(CiteClass):
    """A wrapper for the dust_extinction package.

    References
    ----------
    Gordon 2024, JOSS, 9(100), 7023.
    https://github.com/karllark/dust_extinction

    Attributes
    ----------
    ext_obj : object
        The extinction object from the dust_extinction library. This will
        have r_v set during creation.

    Parameters
    ----------
    model_name : str
        The name of the extinction object in the dust_extinction library.
    r_v : float, optional
        The extinction parameter R(V). Optional for some models.
    **kwargs : `dict`, optional
        Any additional keyword arguments.
    """

    def __init__(self, model_name=None, r_v=None, **kwargs):
        try:
            import dust_extinction  # noqa: F401
        except ImportError as err:  # pragma: no cover
            raise ImportError(
                "The dust_extinction package is needed to use the DustExtinctionEffect. Please install it via"
                "`pip install dust_extinction` or `conda install conda-forge::dust_extinction`."
            ) from err

        # We scan all of the submodules in the dust_extinction package,
        # looking for a matching name.
        self.ext_obj = None
        for submodule in iter_modules(dust_extinction.__path__):
            ext_module = importlib.import_module(f"dust_extinction.{submodule.name}")
            if ext_module is not None and model_name in dir(ext_module):
                ext_class = getattr(ext_module, model_name)

                # dust_extinction models optionally take Rv as a parameter during creation.
                self.ext_obj = ext_class(Rv=r_v, **kwargs)
                break

        if self.ext_obj is None:
            supported_models = self.list_extinction_models()
            raise KeyError(
                f"Invalid dust extinction model '{model_name}'. Supported models are: {supported_models}. "
            )

    @staticmethod
    def list_extinction_models():
        """List the extinction models from the dust_extinction package
        (https://github.com/karllark/dust_extinction)

        Returns
        -------
        list of str
            A list of the names of the extinction models.
        """
        model_names = []

        try:
            import dust_extinction  # noqa: F401
        except ImportError as err:  # pragma: no cover
            raise ImportError(
                "The dust_extinction package is needed to use the DustExtinctionEffect. Please install it via"
                "`pip install dust_extinction` or `conda install conda-forge::dust_extinction`."
            ) from err

        # We scan all of the submodules in the dust_extinction package,
        # looking for classes with extinguish() functions.
        for submodule in iter_modules(dust_extinction.__path__):
            ext_module = importlib.import_module(f"dust_extinction.{submodule.name}")
            for entry_name in dir(ext_module):
                entry_obj = getattr(ext_module, entry_name)
                if hasattr(entry_obj, "extinguish"):
                    model_names.append(entry_name)
        return model_names

    def apply(self, flux_density, times=None, wavelengths=None, ebv=None, **kwargs):
        """Apply the extinction effect to the flux density.

        Parameters
        ----------
        flux_density : numpy.ndarray
            A length T X N matrix of flux density values (in nJy).
        times : numpy.ndarray, optional
            A length T array of times (in MJD). Not used for this effect.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        ebv : float, optional
            The extinction parameter E(B-V). Raises an error if None is provided.
        **kwargs : `dict`, optional
           Any additional keyword arguments, including any additional
           parameters needed to apply the effect.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of flux densities after the effect is applied (in nJy).
        """
        if ebv is None:
            raise ValueError("ebv must be provided")
        if wavelengths is None:
            raise ValueError("wavelengths must be provided")

        # The extinction factor computed by dust_extinction is a multiplicative
        # factor to reduce the flux (<= 1 for all wavelengths).
        ext_factor = self.ext_obj.extinguish(wavelengths * u.angstrom, Ebv=ebv)
        return flux_density * ext_factor


class _SncosmoExtinctionWrapper(CiteClass):
    """A wrapper for the (sncosmo) extinction package.

    References
    ----------
    https://github.com/sncosmo/extinction
    Barbary et. al. 2017
    doi: 10.5281/zenodo.804966

    Attributes
    ----------
    model_name : str
        The model_name of the extinction model to use.
    extinction_fn : function
        The extinction function from the backend library.

    Parameters
    ----------
    model_name : str
        The name of the extinction object in the extinction library.
    r_v : float, optional
        The extinction parameter R(V). Optional for some models.
    **kwargs : `dict`, optional
        Any additional keyword arguments.
    """

    def __init__(self, model_name=None, r_v=None, **kwargs):
        try:
            import extinction  # noqa: F401
        except ImportError as err:  # pragma: no cover
            raise ImportError(
                "The extinction package is needed to use the ExtinctionEffect. Please install it via"
                "`pip install extinction` or `conda install conda-forge::extinction`."
            ) from err

        # We scan all of the submodules in the extinction package,
        # looking for a matching name.
        if model_name not in dir(extinction):
            raise KeyError(f"Invalid extinction model '{model_name}' from extinction package.")
        self.extinction_fn = getattr(extinction, model_name)
        self.model_name = model_name

        # Check that we either have r_v set or that the model does not need it.
        if model_name == "fm07":
            if r_v is not None and np.abs(r_v - 3.1) > 1e-8:
                raise ValueError("The fm07 extinction model requires r_v to be fixed at 3.1.")
            self.r_v = 3.1
        elif r_v is not None and np.isscalar(r_v):
            self.r_v = float(r_v)
        else:  # pragma: no cover
            raise ValueError(f"The {model_name} extinction model requires a numeric r_v parameter.")

    def apply(
        self,
        flux_density,
        times=None,
        wavelengths=None,
        ebv=None,
        **kwargs,
    ):
        """Apply the extinction effect to the flux density.

        Parameters
        ----------
        flux_density : numpy.ndarray
            A length T X N matrix of flux density values (in nJy).
        times : numpy.ndarray, optional
            A length T array of times (in MJD). Not used for this effect.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        ebv : float, optional
            The extinction parameter E(B-V). Raises an error if None is provided.
        **kwargs : `dict`, optional
           Any additional keyword arguments, including any additional
           parameters needed to apply the effect.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of flux densities after the effect is applied (in nJy).
        """
        if ebv is None:
            raise ValueError("ebv must be provided")
        if wavelengths is None:
            raise ValueError("wavelengths must be provided")

        # Compute A(V) from E(B-V) and R(V)
        a_v = ebv * self.r_v

        # Use the function to compute the extinction magnitudes. Note that the fm07 model
        # requires only a_v (r_v is fixed at 3.1).
        if self.model_name == "fm07":
            ext_mags = self.extinction_fn(wavelengths, a_v=a_v, unit="aa")
        else:
            ext_mags = self.extinction_fn(wavelengths, a_v=a_v, r_v=self.r_v, unit="aa")

        # Apply the extinction to the flux density.
        ext_factor = 10 ** (-0.4 * ext_mags)  # Convert mags to flux factor
        return flux_density * ext_factor[np.newaxis, :]  # Broadcast to all times


class ExtinctionEffect(EffectModel):
    """A general extinction effect model that supports multiple backend libraries.

    Attributes
    ----------
    model_name : str
        The model_name of the extinction model to use.
    ebv : parameter
        The setter (function) for the extinction parameter E(B-V).
    frame : str
        The frame for extinction. 'rest' or 'observer'.
    r_v : float, optional
        The value for the extinction parameter R(V) if needed by the backend model.
    backend : str
        The backend extinction library to use. One of 'dust_extinction' or 'extinction'.
        Default is 'dust_extinction'.

    Parameters
    ----------
    extinction_model : str
        The name of the extinction model to use.
    ebv : parameter
        The setter (function) for the extinction parameter E(B-V).
    r_v : float, optional
        The value for the extinction parameter R(V) if needed by the backend model.
    frame : str
        The frame for extinction. 'rest' or 'observer'.
    backend : str
        The backend extinction library to use. One of 'dust_extinction' or 'extinction'.
    **kwargs : `dict`, optional
        Any additional keyword arguments.
    """

    _BACKEND_TO_WRAPPER = {
        "dust_extinction": _DustExtinctionWrapper,
        "extinction": _SncosmoExtinctionWrapper,
    }

    def __init__(
        self,
        extinction_model=None,
        *,
        ebv=None,
        frame=None,
        r_v=None,
        backend=None,
        **kwargs,
    ):
        self.model_name = extinction_model
        self.backend = backend
        self.r_v = r_v

        # Set up the effect parameters (ebv).
        super().__init__(**kwargs)
        self.add_effect_parameter("ebv", ebv)

        # Set the frame.
        if frame == "observer":
            self.rest_frame = False
        elif frame == "rest":
            self.rest_frame = True
        else:
            raise ValueError("frame must be 'observer' or 'rest'.")

        # Check the backend is one of the supported ones and is installed.
        if self.backend is None or self.backend not in self._BACKEND_TO_WRAPPER:
            raise ValueError(f"backend must be one of {list(self._BACKEND_TO_WRAPPER.keys())}")
        if importlib.util.find_spec(self.backend) is None:  # pragma: no cover
            raise ImportError(
                f"The specified backend '{self.backend}' library is not installed. Please install it via"
                f"`pip install {self.backend}` or `conda install conda-forge::{self.backend}`."
            )

        # Load the correct extinction model.
        self._extinction_wrapper = self._BACKEND_TO_WRAPPER[self.backend](self.model_name, r_v=self.r_v)

    def __repr__(self):
        return f"ExtinctionEffect({self.backend}.{self.model_name}, r_v={self.r_v})"

    def __getstate__(self):
        """We override the default pickling behavior to handle the extinction model, since
        it may not be picklable.
        """
        if self.model_name is None or self.backend is None:  # pragma: no cover
            raise ValueError(
                "Both the model name and backend must be specified (as strings) in order to "
                "to be pickled and used with distributed computation."
            )

        # Return the state without the extinction model, since it may not be picklable.
        state = self.__dict__.copy()
        del state["_extinction_wrapper"]
        return state

    def __setstate__(self, state):
        """We override the default unpickling behavior to handle the extinction model, since
        it may not be picklable.
        """
        self.__dict__.update(state)
        if self.model_name is None or self.backend is None:  # pragma: no cover
            raise ValueError(
                "Both the model name and backend must be specified (as strings) in order to "
                "to be pickled and used with distributed computation."
            )

        self._extinction_wrapper = self._BACKEND_TO_WRAPPER[self.backend](self.model_name, r_v=self.r_v)

    def apply(self, flux_density, times=None, wavelengths=None, ebv=None, **kwargs):
        """Apply the extinction effect to the flux density.

        Parameters
        ----------
        flux_density : numpy.ndarray
            A length T X N matrix of flux density values (in nJy).
        times : numpy.ndarray, optional
            A length T array of times (in MJD). Not used for this effect.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        ebv : float, optional
            The extinction parameter E(B-V). Raises an error if None is provided.
        **kwargs : `dict`, optional
           Any additional keyword arguments, including any additional
           parameters needed to apply the effect.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of flux densities after the effect is applied (in nJy).
        """
        return self._extinction_wrapper.apply(
            flux_density,
            times=times,
            wavelengths=wavelengths,
            ebv=ebv,
            **kwargs,
        )
